###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""ProductionRequestHandler is the implementation of the Production Request
service."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re
import tempfile
import six

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Core.Utilities.Subprocess import shellCall

from LHCbDIRAC.ProductionManagementSystem.DB.ProductionRequestDB import ProductionRequestDB


class ProductionRequestHandler(RequestHandler):

  @classmethod
  def initializeHandler(cls, serviceInfoDict):
    cls.productionRequestDB = ProductionRequestDB()
    return S_OK()

  def __clientCredentials(self):
    creds = self.getRemoteCredentials()
    group = creds.get('group', '(unknown)')
    DN = creds.get('DN', '(unknown)')
#    if 'DN' in creds:
#      cn = re.search('/CN=([^/]+)',creds['DN'])
#      if cn:
#        return { 'User':cn.group(1), 'Group':group }
    return {'User': creds.get('username', 'Anonymous'),
            'Group': group,
            'DN': DN}

  types_createProductionRequest = [dict]

  def export_createProductionRequest(self, requestDict):
    """Create production request."""
    creds = self.__clientCredentials()
    if 'MasterID' not in requestDict:
      requestDict['RequestAuthor'] = creds['User']
    return self.productionRequestDB.createProductionRequest(requestDict, creds)

  types_getProductionRequest = [list]

  def export_getProductionRequest(self, requestIDList):
    """ Get production request(s) specified by the list of requestIDs
        AZ!!: not tested !!
    """
    if not requestIDList:
      return S_OK({})
    result = self.productionRequestDB.getProductionRequest(requestIDList)
    if not result['OK']:
      return result
    rows = {}
    for row in result['Value']['Rows']:
      iD = row['RequestID']
      rows[iD] = row
    return S_OK(rows)

  types_getProductionRequestList = [six.integer_types, six.string_types, six.string_types,
                                    six.integer_types, six.integer_types, dict]

  def export_getProductionRequestList(self, subrequestFor, sortBy, sortOrder, offset, limit, rFilter):
    """Get production requests in list format (for portal grid)"""
    return self.productionRequestDB.getProductionRequest(
        [], subrequestFor, sortBy, sortOrder, offset, limit, rFilter)

  types_updateProductionRequest = [six.integer_types, dict]

  def export_updateProductionRequest(self, requestID, requestDict):
    """Update production request specified by requestID."""
    creds = self.__clientCredentials()
    return self.productionRequestDB.updateProductionRequest(requestID, requestDict, creds)

  types_duplicateProductionRequest = [six.integer_types, bool]

  def export_duplicateProductionRequest(self, requestID, clearpp):
    """Duplicate production request with subrequests."""
    creds = self.__clientCredentials()
    return self.productionRequestDB.duplicateProductionRequest(requestID, creds, clearpp)

  types_deleteProductionRequest = [six.integer_types]

  def export_deleteProductionRequest(self, requestID):
    """Delete production request specified by requestID."""
    creds = self.__clientCredentials()
    return self.productionRequestDB.deleteProductionRequest(requestID, creds)

  types_splitProductionRequest = [six.integer_types, list]

  def export_splitProductionRequest(self, requestID, splitList):
    """split production request."""
    creds = self.__clientCredentials()
    return self.productionRequestDB.splitProductionRequest(requestID, splitList, creds)

  types_getProductionProgressList = [six.integer_types]

  def export_getProductionProgressList(self, requestID):
    """Return the list of associated with requestID productions."""
    return self.productionRequestDB.getProductionProgress(requestID)

  types_addProductionToRequest = [dict]

  def export_addProductionToRequest(self, pdict):
    """Associate production to request."""
    return self.productionRequestDB.addProductionToRequest(pdict)

  types_removeProductionFromRequest = [six.integer_types]

  def export_removeProductionFromRequest(self, productionID):
    """Deassociate production."""
    return self.productionRequestDB.removeProductionFromRequest(productionID)

  types_useProductionForRequest = [six.integer_types, bool]

  def export_useProductionForRequest(self, productionID, used):
    """Set Used flags for production."""
    return self.productionRequestDB.useProductionForRequest(productionID, used)

  types_getRequestHistory = [six.integer_types]

  def export_getRequestHistory(self, requestID):
    """Return the list of state changes for the request."""
    return self.productionRequestDB.getRequestHistory(requestID)

  types_getTrackedProductions = []

  def export_getTrackedProductions(self):
    """Return the list of productions in active requests."""
    return self.productionRequestDB.getTrackedProductions()

  types_updateTrackedProductions = [list]

  def export_updateTrackedProductions(self, update):
    """Update tracked productions (used by Agent)"""
    return self.productionRequestDB.updateTrackedProductions(update)

  types_getTrackedInput = []

  def export_getTrackedInput(self):
    """Return the list of requests with dynamic input data."""
    return self.productionRequestDB.getTrackedInput()

  types_updateTrackedInput = [list]

  def export_updateTrackedInput(self, update):
    """Update real number of input events (used by Agent)"""
    return self.productionRequestDB.updateTrackedInput(update)

  types_getAllSubRequestSummary = []

  def export_getAllSubRequestSummary(self, status='', rType=''):
    """Return a summary for each subrequest."""
    return self.productionRequestDB.getAllSubRequestSummary(status, rType)

  types_getAllProductionProgress = []

  def export_getAllProductionProgress(self):
    """Return all the production progress."""
    return self.productionRequestDB.getAllProductionProgress()

  @staticmethod
  def __getTplFolder(tt):

    csS = PathFinder.getServiceSection('ProductionManagement/ProductionRequest')
    if not csS:
      return S_ERROR("No ProductionRequest parameters in CS")
    tplFolder = gConfig.getValue('%s/templateFolder' % csS, '')
    if not tplFolder:
      return S_ERROR("No templateFolder in ProductionRequest parameters in CS")
    if not os.path.exists(tplFolder) or not os.path.isdir(tplFolder):
      return S_ERROR("Template Folder %s doesn't exist" % tplFolder)
    return S_OK(tplFolder)

  def __getTemplate(self, tt, name):
    ret = self.__getTplFolder(tt)
    if not ret['OK']:
      return ret
    tplFolder = ret['Value']
    if not os.path.exists(os.path.join(tplFolder, name)):
      return S_ERROR("Template %s doesn't exist" % name)
    try:
      with open(os.path.join(tplFolder, name)) as f:
        body = f.read()
    except OSError as e:
      return S_ERROR("Can't read template", str(e))
    return S_OK(body)

  def __productionTemplateList(self, tt):
    """Return production template list (file based)"""
    ret = self.__getTplFolder(tt)
    if not ret['OK']:
      return ret
    tplFolder = ret['Value']
    tpls = [x for x in os.listdir(tplFolder)
            if os.path.isfile(os.path.join(tplFolder, x))]
    results = []
    for tpl in tpls:
      if tpl[-1] == '~':
        continue
      result = self.__getTemplate(tt, tpl)
      if not result['OK']:
        return result
      body = result['Value']
      rcsid = re.search("__RCSID__ = \"([^$]*)\"", body)
      ptime = ''
      author = ''
      ver = ''
      if rcsid:
        # the following line tries to extract author, publishing time, and version
        rcsid = re.match(r"([^ ]+) \((.*)\) (.*)", rcsid.group(1))
        if rcsid:
          ptime = rcsid.group(2)
          author = rcsid.group(3)
          ver = rcsid.group(1)
          tpl = {"AuthorGroup": '',
                 "Author": author,
                 "PublishingTime": ptime,
                 "LongDescription": '',
                 "WFName": tpl,
                 "AuthorDN": '',
                 "WFParent": '',
                 "Description": ver}
          results.append(tpl)
    return S_OK(results)

  types_getProductionTemplateList = []

  def export_getProductionTemplateList(self):
    """Return production template list (file based)"""
    return self.__productionTemplateList('template')

  types_getProductionTemplate = [six.string_types]

  def export_getProductionTemplate(self, name):
    return self.__getTemplate('template', name)

  types_execProductionScript = [six.string_types, six.string_types]

  def export_execProductionScript(self, script, workflow):
    creds = self.__clientCredentials()
    if creds['Group'] != 'lhcb_prmgr':
      return S_ERROR("You have to be production manager")
    result = gProxyManager.downloadProxyToFile(creds['DN'], creds['Group'],
                                               filePath=False,
                                               requiredTimeLeft=86400,
                                               cacheTime=86400)
    if not result['OK']:
      return result
    proxyFile = result['Value']

    try:
      f = tempfile.mkstemp()
      os.write(f[0], workflow)
      os.close(f[0])
      fs = tempfile.mkstemp()
      os.write(fs[0], script)
      os.close(fs[0])
    except OSError as msg:
      gLogger.error("In temporary files creation: " + str(msg))
      os.remove(proxyFile)
      return S_ERROR(str(msg))
    setenv = "source /opt/dirac/bashrc"
    proxy = "X509_USER_PROXY=%s" % proxyFile
    cmd = "python %s %s" % (fs[1], f[1])
    try:
      res = shellCall(1800, ["/bin/bash -c '%s;%s %s'" % (setenv, proxy, cmd)])
      if res['OK']:
        result = S_OK(str(res['Value'][1]) + str(res['Value'][2]))
      else:
        gLogger.error(res['Message'])
        result = res
    except Exception as msg:  # pylint: disable=broad-except
      gLogger.error("During execution: " + str(msg))
      result = S_ERROR("Failed to execute: %s" % str(msg))
    os.remove(f[1])
    os.remove(fs[1])
    os.remove(proxyFile)
    return result

  types_execWizardScript = [six.string_types, dict]

  def export_execWizardScript(self, wizard, wizpar):
    """Execure wizard with parameters."""
    creds = self.__clientCredentials()
    if creds['Group'] != 'lhcb_prmgr':
      # return S_ERROR("You have to be production manager")
      if 'Generate' in wizpar:
        del wizpar['Generate']
    result = gProxyManager.downloadProxyToFile(creds['DN'], creds['Group'],
                                               filePath=False,
                                               requiredTimeLeft=86400,
                                               cacheTime=86400)
    if not result['OK']:
      return result
    proxyFile = result['Value']['proxyFile']
    try:
      f = tempfile.mkstemp()
      os.write(f[0], "wizardParameters = {\n")
      for name, value in wizpar.items():
        os.write(f[0], "  \"" + str(name) + "\": \"\"\"" + str(value) + "\"\"\",\n")
      os.write(f[0], "}\n")
      os.write(f[0], wizard)
      os.close(f[0])
    except Exception as msg:
      gLogger.error("In temporary files createion: " + str(msg))
      os.remove(proxyFile)
      return S_ERROR(str(msg))
    setenv = "source /opt/dirac/bashrc"
    # #proxy = "X509_USER_PROXY=xxx"
    proxy = "X509_USER_PROXY=%s" % proxyFile
    cmd = "python %s" % (f[1])
    try:
      res = shellCall(1800, ["/bin/bash -c '%s;%s %s'"
                             % (setenv, proxy, cmd)])
      if res['OK']:
        result = S_OK(str(res['Value'][1]) + str(res['Value'][2]))
      else:
        gLogger.error(res['Message'])
        result = res
    except Exception as msg:
      gLogger.error("During execution: " + str(msg))
      result = S_ERROR("Failed to execute: %s" % str(msg))
    os.remove(f[1])
    os.remove(proxyFile)
    return result

  types_getProductionList = [six.integer_types]

  def export_getProductionList(self, requestID):
    """Return the list of productions associated with request and its
    subrequests."""
    return self.productionRequestDB.getProductionList(requestID)

  types_getProductionRequestSummary = [[six.string_types, list], [six.string_types, list]]

  def export_getProductionRequestSummary(self, status, requestType):
    """Method to retrieve the production / request relations for a given
    request status."""
    if isinstance(requestType, six.string_types):
      reqTypes = [requestType]
    elif isinstance(requestType, list):
      reqTypes = requestType

    if isinstance(status, six.string_types):
      selectStatus = [status]
    elif isinstance(status, list):
      selectStatus = status

    reqList = self.productionRequestDB.getProductionRequest([])
    if not reqList['OK']:
      return reqList

    requests = reqList['Value']
    resultDict = {}

    for req in requests['Rows']:
      iD = int(req['RequestID'])
      if not req['RequestType'] in reqTypes:
        gLogger.verbose('Skipping %s request ID %s...' % (req['RequestType'], iD))
        continue
      if not req['RequestState'] in selectStatus:
        gLogger.verbose('Skipping request ID %s in state %s' % (iD, req['RequestState']))
        continue
      if req['HasSubrequest']:
        gLogger.verbose('Simulation request %s is a parent, getting subrequests...' % iD)
        subReq = self.productionRequestDB.getProductionRequest([], int(iD))
        if not subReq['OK']:
          gLogger.error('Could not get production request for %s' % iD)
          return subReq
        for sreq in subReq['Value']['Rows']:
          sid = int(sreq['RequestID'])
          resultDict[sid] = {'reqTotal': sreq['rqTotal'],
                             'bkTotal': sreq['bkTotal'],
                             'master': iD}
      else:
        gLogger.verbose('Simulation request %s is a single request' % iD)
        resultDict[iD] = {'reqTotal': req['rqTotal'],
                          'bkTotal': req['bkTotal'],
                          'master': 0}

    return S_OK(resultDict)

  types_getFilterOptions = []

  def export_getFilterOptions(self):
    """Return the dictionary with possible values for filter."""
    return self.productionRequestDB.getFilterOptions()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_PMS_Client_ProcessingProgress."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

__RCSID__ = "$Id$"

dummyResults = {}


class DummyReturn():

  def __init__(self, *args, **kwargs):
    pass

  def __getattr__(self, name):
    return self.dummyMethod

  def dummyMethod(self, *args, **kwargs):
    return dummyResults[self.__class__.__name__]


class dTable(DummyReturn):
  pass

################################################################################


class HTMLProgressTable_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""

    # We need the proper software, and then we overwrite it.
    import LHCbDIRAC.ProductionManagementSystem.Client.ProcessingProgress as moduleTested

    self.progress = moduleTested.HTMLProgressTable

  def tearDown(self):
    """TearDown."""
    del self.progress

################################################################################


class HTMLProgressTable_Success(HTMLProgressTable_TestCase):

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""

    global dummyResults
    dummyResults['dTable'] = None

    progress = self.progress('processingPass')
    self.assertEqual('HTMLProgressTable', progress.__class__.__name__)

################################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
 Test a plugin
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


class FakeClient(object):
  def __init__(self, trans, transID, lfns, asIfProd):
    self.trans = trans
    self.transID = transID
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    self.transClient = TransformationClient()
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    self.bk = BookkeepingClient()
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    self.dm = DataManager()
    self.asIfProd = asIfProd

    (self.transFiles, self.transReplicas) = self.prepareForPlugin(lfns)

  def addFilesToTransformation(self, transID, lfns):
    return S_OK({'Failed': {}, 'Successful': dict([(lfn, 'Added') for lfn in lfns])})

  def getTransformation(self, transID, extraParams=False):
    if transID == self.transID and self.asIfProd:
      transID = self.asIfProd
    if transID != self.transID:
      return self.transClient.getTransformation(transID)
    res = self.trans.getType()
    return DIRAC.S_OK({'Type': res['Value']})

  def getReplicas(self):
    return self.transReplicas

  def getFiles(self):
    return self.transFiles

  def getCounters(self, table, attrList, condDict):
    if condDict['TransformationID'] == self.transID and self.asIfProd:
      condDict['TransformationID'] = self.asIfProd
    if condDict['TransformationID'] != self.transID:
      return self.transClient.getCounters(table, attrList, condDict)
    possibleTargets = ['CERN-RAW', 'CNAF-RAW', 'GRIDKA-RAW', 'IN2P3-RAW',
                       'SARA-RAW', 'PIC-RAW', 'RAL-RAW', 'RRCKI-RAW']
    counters = []
    for se in possibleTargets:
      counters.append(({'UsedSE': se}, 0))
    return DIRAC.S_OK(counters)

  def getBookkeepingQuery(self, transID):
    if transID == self.transID and self.asIfProd:
      return self.transClient.getBookkeepingQuery(self.asIfProd)
    return self.trans.getBkQuery()

  def insertTransformationRun(self, transID, runID, xx):
    return DIRAC.S_OK()

  def getTransformationRuns(self, condDict):
    if condDict['TransformationID'] == self.transID and self.asIfProd:
      condDict['TransformationID'] = self.asIfProd
    if condDict['TransformationID'] == self.transID:
      transRuns = []
      runs = condDict.get('RunNumber', [])
      if not runs and self.transFiles:
        res = self.bk.getFileMetadata([fileDict['LFN'] for fileDict in self.transFiles])
        if not res['OK']:
          return res
        runs = list(set(meta['RunNumber'] for meta in res['Value']['Successful'].values()))
      for run in runs:
        transRuns.append({'RunNumber': run, 'Status': "Active", "SelectedSite": None})
      return DIRAC.S_OK(transRuns)
    else:
      return self.transClient.getTransformationRuns(condDict)

  def getTransformationFiles(self, condDict=None):
    if condDict.get('TransformationID') == self.transID and self.asIfProd:
      condDict['TransformationID'] = self.asIfProd
    if condDict.get('TransformationID') == self.transID:
      transFiles = []
      if 'Status' in condDict and 'Unused' not in condDict['Status']:
        return DIRAC.S_OK(transFiles)
      runs = None
      if 'RunNumber' in condDict:
        runs = condDict['RunNumber']
        if not isinstance(runs, list):
          runs = [runs]
      for fileDict in self.transFiles:
        if not runs or fileDict['RunNumber'] in runs:
          transFiles.append({'LFN': fileDict['LFN'], 'Status': 'Unused', 'RunNumber': fileDict['RunNumber']})
      return DIRAC.S_OK(transFiles)
    else:
      return self.transClient.getTransformationFiles(condDict=condDict)

  def setParameterToTransformationFiles(self, transID, lfnDict):
    """Update the transFiles with some parameters."""
    if transID == self.transID:
      for fileDict in self.transFiles:
        fileDict.update(lfnDict.get(fileDict['LFN'], {}))
      return S_OK()
    else:
      return self.transClient.setParameterToTransformationFiles(transID, lfnDict)

  def getTransformationFilesCount(self, transID, field, selection=None):
    if selection is None:
      selection = {}
    if transID == self.transID or selection.get('TransformationID') == self.transID:
      runs = selection.get('RunNumber')
      if runs and not isinstance(runs, list):
        runs = [runs]
      if field == 'Status':
        counters = {'Unused': 0}
        for fileDict in self.transFiles:
          if not runs or fileDict['RunNumber'] in runs:
            counters['Unused'] += 1
      elif field == 'RunNumber':
        counters = {}
        for fileDict in self.transFiles:
          runID = fileDict['RunNumber']
          if not runs or runID in runs:
            counters.setdefault(runID, 0)
            counters[runID] += 1
      else:
        return DIRAC.S_ERROR('Not implemented for field ' + field)
      counters['Total'] = sum(count for count in counters.values())
      return DIRAC.S_OK(counters)
    else:
      return self.transClient.getTransformationFilesCount(transID, field, selection=selection)

  def getTransformationRunStats(self, transIDs):
    counters = {}
    for transID in transIDs:
      if transID == self.transID:
        for fileDict in self.transFiles:
          runID = fileDict['RunNumber']
          counters[transID][runID]['Unused'] = \
              counters.setdefault(transID, {}).setdefault(runID, {}).setdefault('Unused', 0) + 1
        for runID in counters[transID]:
          counters[transID][runID]['Total'] = counters[transID][runID]['Unused']
      else:
        res = self.transClient.getTransformationRunStats(transIDs)
        if res['OK']:
          counters.update(res['Value'])
        else:
          return res
    return DIRAC.S_OK(counters)

  def addRunsMetadata(self, runID, val):
    return self.transClient.addRunsMetadata(runID, val)

  def getRunsMetadata(self, runID):
    return self.transClient.getRunsMetadata(runID)

  def setTransformationRunStatus(self, transID, runID, status):
    return DIRAC.S_OK()

  def setTransformationRunsSite(self, transID, runID, site):
    return DIRAC.S_OK()

  def setFileStatusForTransformation(self, transID, status, lfns):
    return DIRAC.S_OK()

  def addTransformationRunFiles(self, transID, run, lfns):
    return DIRAC.S_OK()

  def setDestinationForRun(self, runID, site):
    return DIRAC.S_OK()

  def getDestinationForRun(self, runID):
    return self.transClient.getDestinationForRun(runID)

  def prepareForPlugin(self, lfns):
    import time
    print("Preparing the plugin input data (%d files)" % len(lfns))
    type = self.trans.getType()['Value']
    if not lfns:
      return (None, None)
    res = self.bk.getFileMetadata(lfns)
    if res['OK']:
      files = []
      for lfn, metadata in res['Value']['Successful'].items():
        runID = metadata.get('RunNumber', 0)
        runDict = {"RunNumber": runID, "LFN": lfn}
        files.append(runDict)
    else:
      print("Error getting BK metadata", res['Message'])
      return ([], {})
    replicas = {}
    startTime = time.time()
    from DIRAC.Core.Utilities.List import breakListIntoChunks
    for lfnChunk in breakListIntoChunks(lfns, 200):
      # print lfnChunk
      if type.lower() in ("replication", "removal"):
        res = self.dm.getReplicas(lfnChunk, getUrl=False)
      else:
        res = self.dm.getReplicasForJobs(lfnChunk, getUrl=False)
      # print res
      if res['OK']:
        for lfn, ses in res['Value']['Successful'].items():
          if ses:
            replicas[lfn] = sorted(ses)
      else:
        print("Error getting replicas of %d files:" % len(lfns), res['Message'])
    print("Obtained replicas of %d files in %.3f seconds" % (len(lfns), time.time() - startTime))
    return (files, replicas)


def printFinalSEs(transType, location, targets):
  targets = targets.split(',')
  if transType == "Removal":
    remain = []
    for x in location:
      r = ','.join([se for se in x.split(',') if se not in targets])
      remain.append(r)
    print("    Remaining SEs:", remain)
  if transType == "Replication":
    total = []
    for x in location:
      r = x + ',' + ','.join([se for se in targets if se not in x.split(',')])
      total.append(r)
    print("    Final SEs:", total)


@DIRACScript()
def main():
  import DIRAC
  from DIRAC.Core.Base import Script
  from LHCbDIRAC.TransformationSystem.Utilities.PluginScript import PluginScript

  pluginScript = PluginScript()
  pluginScript.registerPluginSwitches()
  pluginScript.registerFileSwitches()

  Script.registerSwitch('', 'AsIfProduction=', '   Production # that this test using as source of information')
  Script.registerSwitch('', 'AllFiles', '   Sets visible = False (useful if files were marked invisible)')
  Script.registerSwitch('', 'NoReplicaFiles', '   Also gets the files without replica (just for BK test)')
  Script.registerSwitch('', 'DefaultOptions', '   Gets from CS the default options for a plugin')

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=True)

  from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
  from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation
  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import getRemovalPlugins, getReplicationPlugins
  from DIRAC import gLogger

  asIfProd = None
  allFiles = False
  noRepFiles = False
  defaultOptions = False
  switches = Script.getUnprocessedSwitches()
  for opt, val in switches:
    if opt == 'AsIfProduction':
      asIfProd = int(val)
    elif opt == 'AllFiles':
      allFiles = True
    elif opt == 'NoReplicaFiles':
      noRepFiles = True
    elif opt == 'DefaultOptions':
      defaultOptions = True
  # print pluginScript.getOptions()
  plugin = pluginScript.getOption('Plugin')

  # Just get and print default options from the CS
  if defaultOptions:
    if not plugin:
      gLogger.fatal("No plugin specified")
      DIRAC.exit(1)
    # Get default options from CS
    res = Operations().getOptionsDict('TransformationPlugins/%s' % plugin)
    if res['OK']:
      gLogger.notice("Parameters for plugin %s (*<param> means it is a generic parameter)" % plugin)
      options = res['Value']
      # Get default options for all plugins
      res = Operations().getOptionsDict('TransformationPlugins')
      if res['OK']:
        allOptions = res['Value']
        for opt in set(allOptions) - set(options):
          if opt in pluginScript.seParameters:
            options['*' + opt] = allOptions[opt]
      # SE options first
      for opt in [opt for opt in sorted(options) if 'SEs' in opt]:
        gLogger.notice("\t%s : %s" % (opt, options[opt]))
      # Other options
      for opt in [opt for opt in sorted(options) if 'SEs' not in opt]:
        gLogger.notice("\t%s : %s" % (opt, options[opt]))
      DIRAC.exit(0)
    else:
      gLogger.error("Plugin has no specific parameters")
      DIRAC.exit(1)

  requestID = pluginScript.getOption('RequestID', 0)
  requestedLFNs = pluginScript.getOption('LFNs')
  # print pluginParams

  # FIXME: can be removed when the subLoggers can do it...
  gLogger.showHeaders()

  gLogger.setLevel('INFO')
  # Create the transformation
  transformation = Transformation()
  transType = ''
  if plugin in getRemovalPlugins():
    transType = "Removal"
  elif plugin in getReplicationPlugins():
    transType = "Replication"
  else:
    transType = "Processing"
  transType = pluginScript.getOption('Type', transType)
  transformation.setType(transType)

  visible = 'Yes'
  if allFiles or not plugin or plugin == "DestroyDataset" or \
          pluginScript.getOption('Productions') or transType == 'Processing':
    visible = 'All'
  bkQueryDict = {}
  checkReplica = True
  if not requestedLFNs:
    bkQuery = pluginScript.getBKQuery(visible=visible)
    if noRepFiles and not plugin:
      # FIXME: this should work but doesn't yet...
      # bkQuery.setOption( 'ReplicaFlag', "ALL" )
      checkReplica = False
    bkQueryDict = bkQuery.getQueryDict()
    if list(bkQueryDict) in ([], ['Visible']):
      print("No BK query was given...")
      Script.showHelp(exitCode=2)

  reqID = pluginScript.getRequestID()
  if not requestID and reqID:
    requestID = reqID

  pluginParams = pluginScript.getPluginParameters()
  pluginSEParams = pluginScript.getPluginSEParameters()
  if pluginSEParams:
    for key, val in pluginSEParams.items():
      res = transformation.setSEParam(key, val)
      if not res['OK']:
        print(res['Message'])
        DIRAC.exit(2)
  if pluginParams:
    for key, val in pluginParams.items():
      res = transformation.setAdditionalParam(key, val)
      if not res['OK']:
        print(res['Message'])
        DIRAC.exit(2)

  print("Transformation type:", transType)
  if requestedLFNs:
    print("%d requested LFNs" % len(requestedLFNs))
  else:
    print("BK Query:", bkQueryDict)
  print("Plugin:", plugin)
  print("Parameters:", pluginParams)
  if pluginSEParams:
    print("SE parameters:", pluginSEParams)
  if requestID:
    print("RequestID:", requestID)
  # get the list of files from BK
  if requestedLFNs:
    lfns = requestedLFNs
  else:
    print("Getting the files from BK")
    lfns = bkQuery.getLFNs(
        printSEUsage=(
            (transType == 'Removal' or not plugin) and
            not pluginScript.getOption('Runs') and
            not pluginScript.getOption('DQFlags')
        ),
        printOutput=checkReplica,
        visible=visible,
    )
    if not checkReplica:
      bkQuery.setOption('ReplicaFlag', "No")
      lfns += bkQuery.getLFNs(printSEUsage=False, printOutput=False, visible=visible)
      print('%d files in directories:' % len(lfns))
      directories = {}
      import os
      for lfn in lfns:
        dd = os.path.dirname(lfn)
        directories[dd] = directories.setdefault(dd, 0) + 1
      for dd in sorted(directories):
        print(dd, directories[dd])
  if len(lfns) == 0:
    print("No files found in BK...Exiting now")
    DIRAC.exit(0)

  if not plugin:
    print("No plugin to be tested...")
    DIRAC.exit(0)

  print("\nNow testing the %s plugin %s" % (transType.lower(), plugin))
  transformation.setPlugin(plugin)
  transformation.setBkQuery(bkQueryDict)

  from LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin
  transID = -9999
  pluginParams['TransformationID'] = transID
  pluginParams['Status'] = "Active"
  pluginParams['Type'] = transType
  # Create a fake transformation client
  fakeClient = FakeClient(transformation, transID, lfns, asIfProd)
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  from DIRAC.DataManagementSystem.Client.DataManager import DataManager
  oplugin = TransformationPlugin(plugin, transClient=fakeClient,
                                 dataManager=DataManager(), bkClient=BookkeepingClient())
  pluginParams['TransformationID'] = transID
  pluginParams.update(pluginSEParams)
  oplugin.setParameters(pluginParams)
  replicas = fakeClient.getReplicas()
  files = fakeClient.getFiles()
  if not replicas:
    print("No replicas were found, exit...")
    DIRAC.exit(2)
  oplugin.setInputData(replicas)
  oplugin.setTransformationFiles(files)
  oplugin.setDebug(pluginScript.getOption('Debug', False))
  import time
  startTime = time.time()
  res = oplugin.run()
  print("Plugin took %.1f seconds" % (time.time() - startTime))
  print("")
  if res['OK']:
    print(len(res['Value']), "tasks created")
    i = 0
    previousTask = {'First': 0, 'SEs': None, 'Location': None, 'Tasks': 0}
    noReplicaLFNs = [lfn for _targetSE, lfnList in res['Value'] for lfn in lfnList if lfn not in replicas]
    if noReplicaLFNs:
      result = DataManager().getReplicas(noReplicaLFNs, getUrl=False)
      if res['OK']:
        replicas.update(result['Value']['Successful'])
    for task in res['Value']:
      i += 1
      location = []
      for lfn in task[1]:
        ll = ','.join(sorted(replicas.get(lfn, ['Unknown'])))
        # print "LFN", lfn, l
        if ll not in location:
          location.append(ll)
      if len(task[1]) == 1:
        # Only 1 file in task
        if previousTask['Tasks']:
          if task[0] == previousTask['SEs'] and location == previousTask['Location']:
            # Accumulate tasks for the same site and with same origin
            previousTask['Tasks'] += 1
            continue
          else:
            # Print out previous tasks
            if previousTask['First'] == i - 1:
              print('%d' % previousTask['First'], '- Target SEs:', previousTask['SEs'], end=' ')
              '- 1 file - Current locations:', previousTask['Location']
            else:
              print('%d:%d (%d tasks)' % (previousTask['First'], i - 1, i - previousTask['First']), end=' ')
              '- Target SEs:', previousTask['SEs'], "- 1 file", " - Current locations:", previousTask['Location']
            printFinalSEs(transType, previousTask['Location'], previousTask['SEs'])
        previousTask = {'First': i, 'SEs': task[0], 'Location': location, 'Tasks': 1}
      else:
        if previousTask['Tasks']:
          print('%d:%d (%d tasks)' % (previousTask['First'], i - 1, i - previousTask['First']), end=' ')
          '- Target SEs:', previousTask['SEs'], "- 1 file", " - Current locations:", previousTask['Location']
          printFinalSEs(transType, previousTask['Location'], previousTask['SEs'])
          previousTask = {'First': 0, 'SEs': None, 'Location': None, 'Tasks': 0}
        print(i, '- Target SEs:', task[0], "- %d files" % len(task[1]), " - Current locations:", location)
        printFinalSEs(transType, location, task[0])
    if previousTask['Tasks']:
      i += 1
      if i - previousTask['First'] == 1:
        print('%d' % previousTask['First'], '- Target SEs:', previousTask['SEs'], "- 1 file", end=' ')
        " - Current locations:", previousTask['Location']
      else:
        print('%d:%d (%d tasks)' % (previousTask['First'], i - 1, i - previousTask['First']), end=' ')
        '- Target SEs:', previousTask['SEs'], "- 1 file", " - Current locations:", previousTask['Location']
      printFinalSEs(transType, previousTask['Location'], previousTask['SEs'])
  else:
    print(res['Message'])
  DIRAC.exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC/ResourceStatusSystem/Agent/LHCbPRProxyAgent.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# First, pythonic stuff
import os
import stat

# Second, DIRAC stuff
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule


AGENT_NAME = 'ResourceStatus/LHCbPRProxyAgent'


class LHCbPRProxyAgent(AgentModule):
  """Class LHCbPRProxyAgent. This agent is in charge of getting a fresh proxy
  from DIRAC. In this case, it sets the proxy as LHCbPR. See
  CS/Operations/Shift/LHCbPR.

  This is used by the LHCbPR framework, to get a new proxy every day,
  without having to use any hardcoded password ( which means, this agent will
  run on the same machine as LHCbPR ).
  """

  def initialize(self):
    """Method executed when the agent is launched. It sets the proxy as LHCbPR,
    which is stored under.

    /opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent
    """

    try:

      self.am_setOption('shifterProxy', 'LHCbPR')
      return S_OK()

    except Exception:
      errorStr = "LHCbPRProxyAgent initialization"
      self.log.exception(errorStr)
      return S_ERROR(errorStr)

  def execute(self):
    """At every execution this method will try to print the environment
    variable X509_USER_PROXY, which should contain the path to the proxy. If is
    empty, something went wrong.

    This agent should get a new proxy whenever it is close to expire.
    Appart from that, which is done internally, it is a pretty dummy
    loop.
    """

    try:
      diracCred = '/opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent/.shifterCred'
      lhcbprCred = '/tmp/x509up_u45054'
      os.environ['X509_USER_PROXY'] = '/opt/dirac/pro/work/ResourceStatus/LHCbPRProxyAgent/.shifterCred'
      import shutil
      shutil.copyfile(diracCred, lhcbprCred)
      os.chmod(lhcbprCred, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP)

      self.log.info('Loop')
      self.log.info(os.environ['X509_USER_PROXY'])
      return S_OK()

    except KeyError as x:
      self.log.exception(x)
      return S_ERROR(x)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It stores the job related information."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobConfiguration import JobConfiguration
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobOption import JobOption
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.File import File
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.InputFile import InputFile
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobParameters import JobParameters
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.Job import Job
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.FileParam import FileParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.SimulationConditions import SimulationConditions
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.DataTakingConditions import DataTakingConditions
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.FileReplica import FileReplica
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Replica.ReplicaParam import ReplicaParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.Quality import Quality
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.QualityParameters import QualityParameters
from DIRAC import gLogger

__RCSID__ = "$Id$"


class JobReader(object):
  """JobReader class."""
  #############################################################################

  def __init__(self):
    pass

  #############################################################################
  def readJob(self, doc, fileName):
    """reads and stores the job elements from the file."""
    job = Job()
    job.setFileName(fileName)
    gLogger.debug("Reading job from" + str(fileName) + "XML!")

    self.__readJobConfigurations(doc, job)
    self.__readJobOptions(doc, job)
    self.__readJobTypeParameters(doc, job)
    self.__readJobInputFiles(doc, job)
    self.__readJobOutputFiles(doc, job)
    self.__readJobSimulationConditions(doc, job)
    self.__readJobDataTakingConditions(doc, job)

    gLogger.debug("Job reading finished succesfully!")
    return job

  #############################################################################
  @staticmethod
  def __readJobConfigurations(doc, job):
    """reads and stores the job configurations."""

    conf = JobConfiguration()
    jobElements = doc.getElementsByTagName('Job')
    for node in jobElements:

      name = node.getAttributeNode('ConfigName')
      if name is not None:
        conf.setConfigName(name.value)
      else:
        gLogger.error("<ConfigName> XML tag is missing!!")

      version = node.getAttributeNode('ConfigVersion')
      if version is not None:
        conf.setConfigVersion(version.value)
      else:
        gLogger.error("<ConfigVersion> XML tag is missing!!")

      date = node.getAttributeNode('Date')
      if date is not None:
        conf.setDate(date.value)
      else:
        gLogger.error("<Date> XML tag is missing!!")

      time = node.getAttributeNode('Time')
      if time is not None:
        conf.setTime(time.value)
      else:
        gLogger.error("<Time> XML tag is missing!!")

    job.setJobConfiguration(conf)

  #############################################################################
  @staticmethod
  def __readJobOptions(doc, job):
    """reads and stores the job option."""
    jobOptions = doc.getElementsByTagName('JobOption')
    for node in jobOptions:
      options = JobOption()
      recipient = node.getAttributeNode('Recipient')
      if recipient is not None:
        options.setRecipient(recipient.value)
      else:
        gLogger.warn("<Recipinet> jobOption XML tag is missing!!")

      name = node.getAttributeNode('Name')
      if name is not None:
        options.setName(name.value)
      else:
        gLogger.warn("<Name> JobOption XML tag is missing!!")

      value = node.getAttributeNode('Value')
      if value is not None:
        options.setValue(value.value)
      else:
        gLogger.warn("<Value> JobOption XML tag is missing!!")

      job.addJobOptions(options)

  #############################################################################
  @staticmethod
  def __readJobTypeParameters(doc, job):
    """reads and stores the job parameters."""

    jobTypeParameters = doc.getElementsByTagName('TypedParameter')

    for node in jobTypeParameters:
      parameters = JobParameters()
      name = node.getAttributeNode('Name')
      if name is not None:
        parameters.setName(name.value)
      else:
        gLogger.warn("<Name> TypedParameter XML tag is missing!!")

      value = node.getAttributeNode('Value')
      if value is not None:
        parameters.setValue(value.value)
      else:
        gLogger.warn("<Value> TypedParameter XML tag is missing!!")

      nodetype = node.getAttributeNode('Type')
      if nodetype is not None:
        parameters.setType(nodetype.value)
      else:
        gLogger.warn("<Type> TypedParameter XML tag is missing!!")

      job.addJobParams(parameters)

  ########################################################################
  @staticmethod
  def __readJobInputFiles(doc, job):
    """reads and stores the input files of a job."""
    jobInputFile = doc.getElementsByTagName('InputFile')
    for node in jobInputFile:
      inputFile = InputFile()
      inputFileName = node.getAttributeNode('Name')
      if inputFileName is not None:
        inputFile.setFileName(inputFileName.value)
      else:
        gLogger.error("<Name> InputFile XML tag is missing!!")

      job.addJobInputFiles(inputFile)

  ########################################################################
  def __readJobOutputFiles(self, doc, job):
    """reads the jobs output files."""
    jobOutputFiles = doc.getElementsByTagName('OutputFile')
    for node in jobOutputFiles:
      outputFile = File()

      name = node.getAttributeNode('Name')
      if name is not None:
        outputFile.setFileName(name.value)
      else:
        gLogger.error("<Name> Outputfile XML tag is missing!!")

      typeName = node.getAttributeNode('TypeName')
      if typeName is not None:
        outputFile.setFileType(typeName.value)
      else:
        gLogger.error("<Type> outputfile XML tag is missing!!")

      version = node.getAttributeNode('TypeVersion')
      if version is not None:
        outputFile.setFileVersion(version.value)
      else:
        gLogger.error("<Version> outputfile XML tag is missing!!")

      fileparams = node.getElementsByTagName('Parameter')
      for param in fileparams:
        outputFileParams = FileParam()

        name = param.getAttributeNode('Name')
        if name is not None:
          outputFileParams.setParamName(name.value)
        else:
          gLogger.error("<Name> outputfile Parameter XML tag is missing!!")

        value = param.getAttributeNode('Value')
        if value is not None:
          outputFileParams.setParamValue(value.value)
        else:
          gLogger.error("<Value> outputfile Parameter XML tag is missing!!")

        outputFile.addFileParam(outputFileParams)
      self.__readFilereplica(doc, outputFile)
      self.__readDataquality(node, outputFile)
      job.addJobOutputFiles(outputFile)

  ########################################################################
  @staticmethod
  def __readFilereplica(doc, outputFile):
    """It creates the replica object and adds to the output file."""
    replicas = doc.getElementsByTagName('Replica')  # I have to check doc ? no node? !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    for replica in replicas:
      rep = FileReplica()
      param = ReplicaParam()
      name = replica.getAttributeNode("Name")
      location = replica.getAttributeNode("Location")
      if name is not None:
        param.setName(name.value)
      if location is not None:
        param.setLocation(location.value)
        rep.addParam(param)
      outputFile.addReplicas(rep)

  ########################################################################
  @staticmethod
  def __readDataquality(node, outputFile):
    """It reads the data quality information of a given file."""
    qualities = node.getElementsByTagName("Quality")
    for quality in qualities:
      fileQuality = Quality()
      group = quality.getAttributeNode("Group")
      if group is None:
        fileQuality.setGroup(group.value)
      else:
        gLogger.warn("<Group> Quality XML tag is missing!!")

      flag = quality.getAttributeNode("Flag")
      if flag is not None:
        fileQuality.setFlag(flag.value)
      else:
        gLogger.warn("<Flag> Quality XML tag is missing!!")

      parameters = quality.getElementsByTagName("Parameter")
      for param in parameters:
        qualityParameters = QualityParameters()
        name = param.getAttributeNode("Name")
        if name is not None:
          qualityParameters.setName(name.value)
        else:
          gLogger.warn("<Name> Quality XML tag is missing!!")

        value = param.getAttributeNode("Value")
        if value is not None:
          qualityParameters.setValue(value.value)
        else:
          gLogger.warn("<Value> Quality XML tag is missing!!")

        fileQuality.addParam(qualityParameters)

      outputFile.addQuality(fileQuality)

  ########################################################################
  @staticmethod
  def __readJobSimulationConditions(doc, job):
    """reads and stores the job simulation condition."""
    gLogger.debug("Read Simulation Conditions")
    simcond = doc.getElementsByTagName('SimulationCondition')
    if len(simcond) != 1:
      gLogger.debug("Too many Simulation conditions:", str(simcond))
    else:
      simParam = SimulationConditions()
      node = simcond[0]
      parameters = node.getElementsByTagName('Parameter')
      for param in parameters:
        name = param.getAttributeNode('Name')
        value = param.getAttributeNode("Value")
        if name is None or value is None:
          gLogger.warn("<Name>  or <Value> simulation XML tag is missing!!")
        else:
          simParam.addParam(name.value, value.value)
          job.addSimulationCond(simParam)

  ########################################################################
  @staticmethod
  def __readJobDataTakingConditions(doc, job):
    """reads and stores the job data taking conditions."""
    gLogger.debug("Read DataTaking Conditions")
    daqcond = doc.getElementsByTagName('DataTakingConditions')
    if len(daqcond) != 1:
      gLogger.debug("Too many DataTaking conditions:", str(daqcond))
    else:
      daqParam = DataTakingConditions()
      node = daqcond[0]
      parameters = node.getElementsByTagName('Parameter')
      for param in parameters:
        name = param.getAttributeNode('Name')
        value = param.getAttributeNode("Value")
        if name is None or value is None:
          gLogger.warn("<Name>  or <Value> DataTakingConditions XML tag is missing!!")
        else:
          daqParam.addParam(name.value, value.value)
          job.addDataTakingCond(daqParam)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Module for creating, describing and managing production requests objects."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import itertools
import copy
import re
import six

from DIRAC import gLogger, S_OK

from DIRAC.Core.Utilities.DAG import DAG
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


class ProductionRequest(object):
  """ Production request class - objects are usually created starting from a production request
  """

  def __init__(self, bkkClientIn=None, diracProdIn=None):
    """c'tor.

    Some variables are defined here. A production request is made of:
    stepsList, productionsTypes, and various parameters of those
    productions
    """

    if bkkClientIn is None:
      self.bkkClient = BookkeepingClient()
    else:
      self.bkkClient = bkkClientIn

    if diracProdIn is None:
      self.diracProduction = DiracProduction()
    else:
      self.diracProduction = diracProdIn

    self.rpcProductionRequest = ProductionRequestClient()
    self.tc = TransformationClient()

    self.logger = gLogger.getSubLogger('ProductionRequest')

    self.opsH = Operations()

    # parameters of the request
    self.requestID = 0
    self.parentRequestID = 0
    self.appendName = '1'
    self.outConfigName = ''
    self.prodsToLaunch = []  # productions to launch
    self.stepsListDict = []  # list of dict of steps
    self.stepsInProds = []  # a list of lists
    self.mcTesting = True
    # parameters of the input data
    self.processingPass = ''
    self.dataTakingConditions = ''
    self.eventType = ''
    self.bkFileType = []
    self.dqFlag = ''
    self.startRun = 1
    self.endRun = 2
    self.runsList = ''
    self.configName = 'test'
    self.configVersion = 'certification'
    # parameters of the first production
    self.publishFlag = True
    self.testFlag = False
    self.derivedProduction = 0
    self.previousProdID = 0  # optional prod from which to start
    self.fullListOfOutputFileTypes = []

    # parameters that are the same for each productions
    self.prodGroup = ''  # This ends up being 'ProcessingType', workflow parameter, and it is used for accounting
    self.visibility = ''  # For BKQuery
    self.fractionToProcess = 0
    self.minFilesToProcess = 0
    self.modulesList = None  # Usually:
    # ['GaudiApplication', 'AnalyseXMLSummary',
    # 'ErrorLogging', 'BookkeepingReport', 'StepAccounting' ]
    self.extraModulesList = None

    # parameters of each production (the length of each list has to be the same as the number of productions
    self.events = []
    self.stepsList = []
    self.extraOptions = {}
    self.prodsTypeList = []
    self.bkQueries = []  # list of bk queries
    self.removeInputsFlags = []
    self.priorities = []
    self.cpus = []
    self.inputs = []  # list of lists
    self.outputModes = []
    self.targets = []
    self.outputFileMasks = []
    self.outputFileSteps = []
    self.groupSizes = []
    self.plugins = []
    self.inputDataPolicies = []
    self.previousProds = [None]  # list of productions from which to take the inputs (the first is always None)
    self.multicore = []  # list of flags to override the multi core flags of the steps
    # list of tuples with min and max number of processors allowed for the jobs ((0,0) = unlimited)
    self.processors = []

    self.outputSEs = []  # a list of StorageElements
    self.specialOutputSEs = []  # a list of dictionaries - might be empty
    self.outputSEsPerFileType = []  # a list of dictionaries - filled later
    self.ancestorDepths = []
    self.compDict = {'HIGH': 'Compression-LZMA-4', 'LOW': 'Compression-ZLIB-1'}
    self.compressionLvl = ['']  # List: one compression level each step
    self.appConfig = '$APPCONFIGOPTS/Persistency/'  # Default location of the compression level configuration files
    #
    # These lists define the visibility of the output files produced by each step. For MC productions, the visibility
    # is tied to compression level. VIsible files are compressed at the highest level
    #
    self.outputVisFlag = []  # List of dictionary with default visibility flag of the output files per single step
    self.specialOutputVisFlag = []  # List of dictionaries with special visibility flag for given file type

  #############################################################################

  def resolveSteps(self):
    """Given a list of steps in strings, some of which might be missing,
    resolve it into a list of dictionary of steps (self.stepsListDict)"""
    outputVisFlag = dict([k, v] for el in self.outputVisFlag for k, v in el.items()
                         )  # Transform the list of dictionaries in a dictionary
    specialOutputVisFlag = dict([k, v] for el in self.specialOutputVisFlag for k, v in el.items())
    count = 0  # Needed to add correctly the optionFiles to the list of dictonaries of steps
    for stepID in self.stepsList:

      stepDict = self.bkkClient.getAvailableSteps({'StepId': stepID})
      if not stepDict['OK']:
        raise ValueError(stepDict['Message'])
      else:
        stepDict = stepDict['Value']

      stepsListDictItem = {}

      s_in = self.bkkClient.getStepInputFiles(stepID)
      if not s_in['OK']:
        raise ValueError(s_in['Message'])
      stepsListDictItem['fileTypesIn'] = [fileType[0].strip() for fileType in s_in['Value']['Records']]

      s_out = self.bkkClient.getStepOutputFiles(stepID)
      if not s_out['OK']:
        raise ValueError(s_out['Message'])
      fileTypesList = [fileType[0].strip() for fileType in s_out['Value']['Records']]
      self.fullListOfOutputFileTypes = self.fullListOfOutputFileTypes + fileTypesList
      stepsListDictItem['fileTypesOut'] = fileTypesList

      for parameter, value in zip(stepDict['ParameterNames'], stepDict['Records'][0]):
        if parameter.lower() in ['conddb', 'dddb', 'dqtag'] and value:
          if value.lower() == 'frompreviousstep':
            value = self.stepsListDict[-1][parameter]

        if parameter == 'OptionFiles':  # Modifying the OptionFiles (for setting the compression level)
          if 'MDF' not in stepsListDictItem['fileTypesOut']:  # certain MC produce MDF, which shouldn't be compressed
            #
            # If the prod manager sets a compression level for a particular step, either we append the option file
            # or we overwrite the existing one inherited with the step
            #
            if len(self.compressionLvl) > count and self.compressionLvl[count] != '':
              persist = re.compile('Compression-[A-Z]{4}-[1-9]')
              # self.compressionLvl[count] = self.appConfig + self.compressionLvl[count] + '.py'
              self.compressionLvl[count] = self.appConfig + self.compDict[self.compressionLvl[count].upper()] + '.py'
              if not persist.search(value):
                if value == '':
                  value = self.compressionLvl[count]
                else:
                  value = ";".join((value, self.compressionLvl[count]))
              else:
                value = persist.sub(persist.search(self.compressionLvl[count]).group(), value)
            #
            # If instead the prod manager doesn't declare a compression level, e.g. for intermediate steps,
            # we check if there is one in the options and in case we delete it. This leaves the default zip level
            # defined inside Gaudi
            #
            elif len(self.compressionLvl) > count and self.compressionLvl[count] == '':
              persist = re.compile(r'\$\w+/Persistency/Compression-[A-Z]{4}-[1-9].py;?')
              if persist.search(value):
                value = persist.sub('', value)

        if parameter == 'SystemConfig' and value is not None and re.search('slc5', value):
          p = re.compile(r'\$\w+/Persistency/Compression-[A-Z]{4}-[1-9].py;?')
          if p.search(stepsListDictItem['OptionFiles']):
            stepsListDictItem['OptionFiles'] = p.sub('', stepsListDictItem['OptionFiles'])

        stepsListDictItem[parameter] = value  # Fixing what decided

      if stepsListDictItem['StepId'] in self.extraOptions:
        stepsListDictItem['ExtraOptions'] = self.extraOptions[stepsListDictItem['StepId']]
      else:
        stepsListDictItem['ExtraOptions'] = ''

      stepsListDictItem['prodStepID'] = str(stepID) + str(stepsListDictItem['fileTypesIn'])

      if 'isMulticore' not in stepsListDictItem:
        stepsListDictItem['isMulticore'] = 'N'

      if 'SystemConfig' not in stepsListDictItem:
        stepsListDictItem['SystemConfig'] = ''

      if 'mcTCK' not in stepsListDictItem:
        stepsListDictItem['mcTCK'] = ''

      # Add visibility info during step resolution
      if 'visibilityFlag' not in stepsListDictItem:
        outputVisList = list({'Visible': outputVisFlag[str(count + 1)],
                              'FileType': ftype} for ftype in stepsListDictItem['fileTypesOut'])
        if str(count + 1) in specialOutputVisFlag:
          for it in outputVisList:
            if it['FileType'] in specialOutputVisFlag[str(count + 1)]:
              it['Visible'] = specialOutputVisFlag[str(count + 1)][it['FileType']]

        stepsListDictItem['visibilityFlag'] = outputVisList

      self.stepsListDict.append(stepsListDictItem)
      count += 1

  #############################################################################

  def buildAndLaunchRequest(self):
    """uses _applyOptionalCorrections, _getProdsDescriptionDict,
    _buildProduction, and DiracProduction.launchProduction."""

    if not self.stepsListDict:
      self.resolveSteps()

    self._applyOptionalCorrections()

    self._determineOutputSEs()

    prodsDict = self._getProdsDescriptionDict()

    stepsListDict = list(self.stepsListDict)

    fromProd = self.previousProdID
    prodsLaunched = []

    self.logger.debug(prodsDict)
    # now we build and launch each productions
    for prodIndex, prodDict in prodsDict.items():

      if self.prodsToLaunch:
        if prodIndex not in self.prodsToLaunch:
          continue

      # build the list of steps in a production
      stepsInProd = []
      for stepID in prodDict['stepsInProd-ProdName']:
        for step in stepsListDict:
          if step['prodStepID'] == stepID:
            stepsInProd.append(stepsListDict.pop(stepsListDict.index(step)))
      # NOT READY (alternative to previous 5 lines)
      # build the DAG of steps in a production
      # stepsInProdDAG = self._getStepsInProdDAG( prodDict, stepsListDict )
      # Here, for today it is just convert to a list
      # TODO: fix this in order to properly use DAGs (now it's only sequential)
      # FIXME: using getIndexNodes we can't assure the order is respected
      # stepsInProd = stepsInProdDAG.getIndexNodes()

      if prodDict['previousProd'] is not None:
        fromProd = prodsLaunched[prodDict['previousProd'] - 1]
        self.previousProdID = fromProd

      prod = self._buildProduction(prodType=prodDict['productionType'],
                                   stepsInProd=stepsInProd,
                                   outputSE=prodDict['outputSE'],
                                   priority=prodDict['priority'],
                                   cpu=prodDict['cpu'],
                                   inputDataList=prodDict['input'],
                                   outputMode=prodDict['outputMode'],
                                   inputDataPolicy=prodDict['inputDataPolicy'],
                                   outputFileMask=prodDict['outputFileMask'],
                                   outputFileStep=prodDict['outputFileStep'],
                                   target=prodDict['target'],
                                   removeInputData=prodDict['removeInputsFlag'],
                                   groupSize=prodDict['groupSize'],
                                   bkQuery=prodDict['bkQuery'],
                                   plugin=prodDict['plugin'],
                                   previousProdID=fromProd,
                                   derivedProdID=prodDict['derivedProduction'],
                                   transformationFamily=prodDict['transformationFamily'],
                                   events=prodDict['events'],
                                   multicore=prodDict['multicore'],
                                   processors=prodDict['processors'],
                                   ancestorDepth=prodDict['ancestorDepth'])

      # if the production is a simulation production type, submit it to the automated testing
      if prodDict['productionType'] in self.opsH.getValue('Transformations/ExtendableTransfTypes', ['MCSimulation']) \
         and self.mcTesting:
        prodID = self._mcSpecialCase(prod, prodDict)

      else:
        res = self.diracProduction.launchProduction(prod=prod,
                                                    publishFlag=self.publishFlag,
                                                    testFlag=self.testFlag,
                                                    requestID=self.requestID,
                                                    tracking=prodDict['tracking'])
        if not res['OK']:
          raise RuntimeError(res['Message'])

        prodID = res['Value']

      prodsLaunched.append(prodID)

      if self.publishFlag:
        self.logger.notice("For request %d, submitted Production" % self.requestID,
                           "%d, of type %s, ID = %s" % (prodIndex,
                                                        prodDict['productionType'],
                                                        str(prodID)))
    return S_OK(prodsLaunched)

  #############################################################################

  def _getStepsInProdDAG(self, prodDict, stepsListDict, stepsOrder='sequential'):
    """Builds the DAG of steps in a production.

    :params dict prodDict: dictionary representing one production
    :params list stepsListDict: list of steps (which are dictionaries) that should be in the production

    :returns: stepsInProd (DAG)
    """
    stepsInProd = DAG()

    inserted = None
    for stepID in prodDict['stepsInProd-ProdName']:
      for step in stepsListDict:
        if step['prodStepID'] == stepID:
          ind = stepsListDict.index(step)
          step = stepsListDict.pop(ind)
          stepsInProd.addNode(step)
          if inserted and stepsOrder == 'sequential':
            stepsInProd.addEdge(inserted, step)
          inserted = step

    return stepsInProd

  def _mcSpecialCase(self, prod, prodDict):
    """Treating the MC special case for putting MC productions in status
    "Testing"."""

    # save the original xml before it is edited for testing
    prod._lastParameters()  # pylint: disable=protected-access

    # launchProduction adds extra parameters, as we 'hot swap' the xml, we
    # need to get these parameters for the un-edited version
    originalProcessingType = prod.prodGroup
    originalPriority = prod.priority

    prodXML = prod.LHCbJob.workflow.toXML()

    prodID = self._modifyAndLaunchMCXML(prod, prodDict)

    # load a production from the original xml to save the priority and processing type
    workflowToSave = fromXMLString(prodXML)
    prod.LHCbJob.workflow = workflowToSave
    prod.setParameter('ProcessingType', 'JDL', str(originalProcessingType), 'ProductionGroupOrType')
    prod.setParameter('Priority', 'JDL', str(originalPriority), 'Job Priority')

    # original xml to save
    descriptionToStore = prod.LHCbJob.workflow.toXML()

    # saving the original xml in the StoredJobDescription table.
    res = self.tc.addStoredJobDescription(prodID, descriptionToStore)
    if not res['OK']:
      self.logger.error("Error calling addStoredJobDescription", res['Message'])
      self.logger.info("Cleaning created production and exiting")
      self.diracProduction.production(res['Value'], 'cleaning')
      raise RuntimeError(res['Message'])

    return prodID

  def _modifyAndLaunchMCXML(self, prod, prodDict):
    """Apply modifications to the workflow XML for MC testing case.

    :param Production prod: Production object
    :param dict prodDict: dictionary with production info

    :returns: res['OK'] or res['ERROR']
    """
    # set the destination and number of events for testing
    destination = self.opsH.getValue("Productions/MCTesting/MCTestingDestination", 'DIRAC.Test.ch')
    numberOfEvents = self.opsH.getValue("Productions/MCTesting/numberOfEvents", '500')
    extendBy = self.opsH.getValue("Productions/MCTesting/extendBy", 20)

    prod.setJobParameters({'Destination': destination})
    prod.LHCbJob.workflow.removeParameter('BannedSites')
    prod.setParameter('numberOfEvents', 'string', str(numberOfEvents), 'Number of events to test')

    # add '1' to the stepMask and add GAUSSHIST to the fileMask
    fileTypesOutLastStep = prod.LHCbJob.workflow.step_instances[-2].findParameter('listoutput').getValue()[
        0]['outputDataType']
    newFileMask = ['GAUSSHIST'] + [ftOut.upper() for ftOut in fileTypesOutLastStep.split(';')]
    stepMaskParameter = prod.LHCbJob.workflow.findParameter('outputDataStep')
    if stepMaskParameter:
      stepMask = stepMaskParameter.getValue().replace(' ', '').split(';')
      newOutputFileStep = ';'.join(sorted(list(set(['1']).union(set(stepMask)))))
    else:
      newOutputFileStep = '1'
    prod.setFileMask(newFileMask, newOutputFileStep)

    # find the file types out already built, append GAUSSHIST and set the new listoutput
    fileTypesOut = prod.LHCbJob.workflow.step_instances[0].findParameter('listoutput').getValue()[0]['outputDataType']
    fileTypesOut = fileTypesOut.split(', ')
    fileTypesOut.append('GAUSSHIST')
    outputFilesList = prod._constructOutputFilesList(fileTypesOut)  # pylint: disable=protected-access
    prod.LHCbJob.workflow.step_instances[0].setValue('listoutput', outputFilesList)

    # increase the priority to 10
    prod.priority = 10

    # launch the test production
    res = self.diracProduction.launchProduction(prod=prod,
                                                publishFlag=self.publishFlag,
                                                testFlag=self.testFlag,
                                                requestID=self.requestID,
                                                extend=extendBy,
                                                tracking=prodDict['tracking'],
                                                MCsimflag=True)
    if not res['OK']:
      self.logger.error("Error launching production", res['Message'])
      raise RuntimeError(res['Message'])

    return res['Value']

  def _determineOutputSEs(self):
    """Fill outputSEsPerFileType based on outputSEs, fullListOfOutputFileTypes
    and specialOutputSEs."""
    for outputSE, specialOutputSEs in zip(self.outputSEs, self.specialOutputSEs):
      outputSEDict = {}
      if not self.fullListOfOutputFileTypes:
        raise ValueError("No steps defined")
      outputSEDict = dict([(fType, outputSE) for fType in self.fullListOfOutputFileTypes])
      if specialOutputSEs:
        outputSEDict.update(specialOutputSEs)
      self.outputSEsPerFileType.append(outputSEDict)

  def _applyOptionalCorrections(self):
    """if needed, calls _splitIntoProductionSteps.

    It also applies other changes
    """
    if len(self.bkQueries) != len(self.prodsTypeList):
      self.bkQueries += ['fromPreviousProd'] * (len(self.prodsTypeList) - len(self.bkQueries))

    if len(self.previousProds) != len(self.prodsTypeList):
      self.previousProds += range(1, len(self.prodsTypeList))

    if len(self.events) != len(self.prodsTypeList):
      self.events += ['-1'] * (len(self.prodsTypeList) - len(self.events))

    if not self.removeInputsFlags:
      removeInputsFlags = []
      for prodType in self.prodsTypeList:
        if prodType.lower() == 'merge':
          removeInputsFlags.append(True)
        else:
          removeInputsFlags.append(False)
      self.removeInputsFlags = removeInputsFlags

    if not self.outputFileMasks:
      self.outputFileMasks = [''] * len(self.prodsTypeList)

    if not self.outputFileSteps:
      self.outputFileSteps = [''] * len(self.prodsTypeList)

    if not self.inputs:
      self.inputs = [[]] * len(self.prodsTypeList)

    if not self.outputModes:
      self.outputModes = ['Any'] * len(self.prodsTypeList)

    if not self.targets:
      self.targets = [''] * len(self.prodsTypeList)

    if not self.inputDataPolicies:
      self.inputDataPolicies = ['download'] * len(self.prodsTypeList)

    if not self.multicore:
      self.multicore = ['True'] * len(self.prodsTypeList)

    if not self.processors:
      self.processors = [(0, 0)] * len(self.prodsTypeList)

    if not self.specialOutputSEs:
      self.specialOutputSEs = [{}] * len(self.prodsTypeList)

    if not self.ancestorDepths:
      self.ancestorDepths = [0] * len(self.prodsTypeList)

    # Checking if we need to split the merging step into many productions
    if 'merge' in [pt.lower() for pt in self.prodsTypeList]:
      i = 0
      indexes = []
      for pt in self.prodsTypeList:
        if pt.lower() == 'merge':
          indexes.append(i)
        i += 1

      for index in indexes:
        # In this case and only in this case I have to split the merging in many productions
        plugin = self.plugins[index]
        outputSE = self.outputSEs[index]
        specialOutputSE = self.specialOutputSEs[index]
        priority = self.priorities[index]
        cpu = self.cpus[index]
        bkQuery = self.bkQueries[index]
        groupSize = self.groupSizes[index]
        preProd = self.previousProds[index]
        removeInputsFlag = self.removeInputsFlags[index]
        outputFileMask = self.outputFileMasks[index]
        outputFileStep = self.outputFileSteps[index]
        inputs = self.inputs[index]
        idp = self.inputDataPolicies[index]
        stepID = self.stepsList[index]
        events = self.events[index]
        targets = self.targets[index]
        multicore = self.multicore[index]
        processors = self.processors[index]
        outputMode = self.outputModes[index]
        ancestorDepth = self.ancestorDepths[index]
        if plugin.lower() != 'byrunfiletypesizewithflush' and 'rootmerging' not in plugin.lower():
          stepToSplit = self.stepsListDict[index]
          numberOfProdsToInsert = len(stepToSplit['fileTypesOut'])
          self.prodsTypeList.remove('Merge')
          self.plugins.pop(index)
          self.outputSEs.pop(index)
          self.specialOutputSEs.pop(index)
          self.priorities.pop(index)
          self.cpus.pop(index)
          self.bkQueries.pop(index)
          self.previousProds.pop(index)
          self.groupSizes.pop(index)
          self.removeInputsFlags.pop(index)
          self.outputFileMasks.pop(index)
          self.outputFileSteps.pop(index)
          self.inputs.pop(index)
          self.inputDataPolicies.pop(index)
          self.stepsList.pop(index)
          self.events.pop(index)
          self.targets.pop(index)
          self.multicore.pop(index)
          self.processors.pop(index)
          self.outputModes.pop(index)
          self.ancestorDepths.pop(index)
          newSteps = _splitIntoProductionSteps(stepToSplit)
          newSteps.reverse()
          self.stepsListDict.remove(stepToSplit)
          last = self.stepsInProds.pop(index)[0]
          for x in range(numberOfProdsToInsert):
            self.prodsTypeList.insert(index, 'Merge')
            self.plugins.insert(index, plugin)
            self.outputSEs.insert(index, outputSE)
            self.specialOutputSEs.insert(index, specialOutputSE)
            self.priorities.insert(index, priority)
            self.cpus.insert(index, cpu)
            self.bkQueries.insert(index, bkQuery)
            self.groupSizes.insert(index, groupSize)
            self.removeInputsFlags.insert(index, removeInputsFlag)
            self.outputFileMasks.insert(index, outputFileMask)
            self.outputFileSteps.insert(index, outputFileStep)
            self.inputs.insert(index, inputs)
            self.inputDataPolicies.insert(index, idp)
            self.stepsList.insert(index, stepID)
            self.previousProds.insert(index, preProd)
            self.stepsListDict.insert(index, newSteps[x])
            self.stepsInProds.insert(index + x, [last + x])
            self.events.insert(index, events)
            self.targets.insert(index, targets)
            self.multicore.insert(index, multicore)
            self.processors.insert(index, processors)
            self.outputModes.insert(index, outputMode)
            self.ancestorDepths.insert(index, ancestorDepth)

    correctedStepsInProds = []
    toInsert = self.stepsInProds[0][0]
    lengths = [len(x) for x in self.stepsInProds]
    for length in lengths:
      li = [toInsert + x for x in range(length)]
      toInsert += length
      correctedStepsInProds.append(li)

    self.stepsInProds = correctedStepsInProds

  #############################################################################

  def _getProdsDescriptionDict(self):
    """Returns a dictionary representing the description of the request (of all
    the productions in it)"""

    prodsDict = {}

    prodNumber = 1

    for prodType, stepsInProd, bkQuery, removeInputsFlag, outputSE, priority, \
        cpu, inputD, outputMode, outFileMask, outFileStep, target, groupSize, plugin, idp, \
        previousProd, events, multicore, processors, ancestorDepth in zip(
            self.prodsTypeList,
            self.stepsInProds,
            self.bkQueries,
            self.removeInputsFlags,
            self.outputSEsPerFileType,
            self.priorities,
            self.cpus,
            self.inputs,
            self.outputModes,
            self.outputFileMasks,
            self.outputFileSteps,
            self.targets,
            self.groupSizes,
            self.plugins,
            self.inputDataPolicies,
            self.previousProds,
            self.events,
            self.multicore,
            self.processors,
            self.ancestorDepths):

      if not self.parentRequestID and self.requestID:
        transformationFamily = self.requestID
      else:
        transformationFamily = self.parentRequestID

      stepsInProdProdNameList = [str(self.stepsList[index - 1]) + str(self.stepsListDict[index - 1]['fileTypesIn'])
                                 for index in stepsInProd]
      prodsDict[prodNumber] = {'productionType': prodType,
                               'stepsInProd': [self.stepsList[index - 1] for index in stepsInProd],
                               'bkQuery': bkQuery,
                               'removeInputsFlag': removeInputsFlag,
                               'tracking': 0,
                               'outputSE': outputSE,
                               'priority': priority,
                               'cpu': cpu,
                               'input': inputD,
                               'outputMode': outputMode,
                               'outputFileMask': outFileMask,
                               'outputFileStep': outFileStep,
                               'target': target,
                               'groupSize': groupSize,
                               'plugin': plugin,
                               'inputDataPolicy': idp,
                               'derivedProduction': 0,
                               'transformationFamily': transformationFamily,
                               'previousProd': previousProd,
                               'stepsInProd-ProdName': stepsInProdProdNameList,
                               'events': events,
                               'multicore': multicore,
                               'processors': processors,
                               'ancestorDepth': ancestorDepth}
      prodNumber += 1

    # tracking the last production(s)
    prodsDict[prodNumber - 1]['tracking'] = 1
    typeOfLastProd = prodsDict[prodNumber - 1]['productionType']
    index = 2
    try:
      while prodsDict[prodNumber - index]['productionType'] == typeOfLastProd:
        prodsDict[prodNumber - index]['tracking'] = 1
        index += 1
    except KeyError:
      pass

    # production derivation, if necessary
    if self.derivedProduction:
      prodsDict[1]['derivedProduction'] = self.derivedProduction

    return prodsDict

  #############################################################################

  def _buildProduction(self, prodType, stepsInProd, outputSE, priority, cpu,
                       inputDataList=None,
                       outputMode='Any',
                       inputDataPolicy='download',
                       outputFileMask='',
                       outputFileStep='',
                       target='',
                       removeInputData=False,
                       groupSize=1,
                       bkQuery=None,
                       plugin='',
                       previousProdID=0,
                       derivedProdID=0,
                       transformationFamily=0,
                       events=-1,
                       multicore='True',
                       processors=(0, 0),
                       ancestorDepth=0):
    """Wrapper around Production API to build a production, given the needed
    parameters.

    Args:
      prodType (str): production type (e.g. 'DataStripping')
      stepsInProd (list): list of steps in the production
      outputSE (dict): dictionary that holds relation between file type and output SE
      priority (int): production priority
      cpu (int): CPU time, in HS06s for jobs in this production

    Returns:
      prod: a Production object
    """
    prod = Production()

    # non optional parameters
    prod.LHCbJob.setType(prodType)
    try:
      fTypeIn = [ft.upper() for ft in stepsInProd[0]['fileTypesIn']]
    except IndexError:
      fTypeIn = []
    prod.LHCbJob.workflow.setName('Request_%d_%s_%s_EventType_%s_%s_%s' % (self.requestID, prodType,
                                                                           self.prodGroup, self.eventType,
                                                                           ''.join([x.split('.')[0] for x in fTypeIn]),
                                                                           self.appendName))
    prod.setBKParameters(configName=self.outConfigName, configVersion=self.configVersion,
                         groupDescriptionOrStepsList=stepsInProd, conditions=self.dataTakingConditions)
    prod.setParameter('eventType', 'string', self.eventType, 'Event Type of the production')
    prod.setParameter('numberOfEvents', 'string', str(events), 'Number of events requested')

    prod.setParameter('multicore', 'string', multicore, 'Flag for enabling gaudi parallel')
    if processors[0] or processors[1]:
      minProcs = processors[0] if processors[0] else None
      maxProcs = processors[1] if processors[1] else None
      prod.LHCbJob.setNumberOfProcessors(minNumberOfProcessors=minProcs,
                                         maxNumberOfProcessors=maxProcs)
    prod.prodGroup = self.prodGroup
    prod.priority = priority
    prod.LHCbJob.workflow.setDescrShort('prodDescription')
    prod.LHCbJob.workflow.setDescription('prodDescription')
    prod.LHCbJob.setCPUTime(cpu)
    prod.plugin = plugin

    # optional parameters
    prod.jobFileGroupSize = groupSize
    if inputDataPolicy:
      prod.LHCbJob.setInputDataPolicy(inputDataPolicy)
    prod.setOutputMode(outputMode)
    if outputFileMask:
      outputFileMask = [m.lower() for m in outputFileMask.replace(' ', '').split(',')]
    if outputFileStep:
      if isinstance(outputFileStep, six.string_types):
        outputFileStep = [m.lower() for m in outputFileStep.replace(' ', '').split(',')]
    prod.setFileMask(outputFileMask, outputFileStep)
    if target:
      if target == 'Tier2':
        prod.banTier1s()
      elif 'BAN' in target:
        sitesToBan = target.split(':')
        if len(sitesToBan) > 1:
          prod.banSites(sitesToBan[1:])
      elif target != 'ALL':
        prod.LHCbJob.setDestination(target)
    prod.LHCbJob.setInputData(inputDataList)
    if ancestorDepth:
      prod.LHCbJob.setAncestorDepth(ancestorDepth)
    if derivedProdID:
      prod.ancestorProduction = derivedProdID
    if transformationFamily:
      prod.transformationFamily = transformationFamily
    if self.fractionToProcess:
      prod.setParameter('FractionToProcess', 'string', str(self.fractionToProcess), 'Fraction to process')
    if self.minFilesToProcess:
      prod.setParameter('MinFilesToProcess', 'string', str(self.minFilesToProcess), 'Min N of Files to process')
    if self.processingPass:
      prod.setParameter('processingPass', 'string', self.processingPass, 'Processing pass of input for the request')

    # Adding optional input BK query
    if bkQuery:
      if bkQuery.lower() == 'full':
        prod.inputBKSelection = self._getBKKQuery()
      elif bkQuery.lower() == 'frompreviousprod':
        prod.inputBKSelection = self._getBKKQuery('frompreviousprod', fTypeIn, previousProdID)

    self.logger.verbose('Launching with BK selection %s' % prod.inputBKSelection)

    prod = self._addStepsToProd(prod, stepsInProd, removeInputData=removeInputData)

    for ft, oSE in outputSE.items():
      prod.outputSEs.setdefault(ft, oSE)

    prod.LHCbJob.setDIRACPlatform()

    return prod

  #############################################################################

  def _addStepsToProd(self, prod, stepsInProd, stepsSequence='sequential', removeInputData=False):
    """Given a Production object, add requested steps (application and
    finalization)

    Args:
      prod (Production): the Production object to which the steps are added
      stepsInProd (DAG): DAG of steps in a production
      stepsSequence (str or dict): applications steps sequence
      removeInputData (bool): flag that indicates if the input data should be removed (for the finalization step)

    Returns:
      prod with steps added
    """
    # Adding the application steps
    firstStep = stepsInProd.pop(0)
    stepName = prod.addApplicationStep(stepDict=firstStep,
                                       modulesList=self.modulesList,
                                       extraModulesList=self.extraModulesList)
    prod.gaudiSteps.append(stepName)

    for step in stepsInProd:
      stepName = prod.addApplicationStep(stepDict=step,
                                         inputData='previousStep',
                                         modulesList=self.modulesList,
                                         extraModulesList=self.extraModulesList)
      prod.gaudiSteps.append(stepName)

    # Adding the finalization step
    if removeInputData:
      prod.addFinalizationStep(['UploadOutputData',
                                'RemoveInputData',
                                'UploadLogFile',
                                'UploadMC',
                                'FailoverRequest'])
    else:
      prod.addFinalizationStep()

    return prod

  def _getBKKQuery(self, mode='full', fileType=None, previousProdID=0):
    """simply creates the bkk query dictionary."""

    if fileType is None:
      fileType = []

    if mode.lower() == 'full':
      bkQuery = {'FileType': ';;;'.join(self.bkFileType),
                 'EventType': str(self.eventType),
                 'ConfigName': self.configName,
                 'ConfigVersion': self.configVersion}

      if self.dataTakingConditions:
        bkQuery['DataTakingConditions'] = self.dataTakingConditions

      if self.processingPass:
        bkQuery['ProcessingPass'] = self.processingPass

      if self.dqFlag:
        bkQuery['DataQualityFlag'] = self.dqFlag.replace(',', ';;;').replace(' ', '')

      if self.startRun and self.runsList or self.endRun and self.runsList:
        raise ValueError("Please don't mix runs list with start/end run")

      if self.endRun and self.startRun:
        if self.endRun < self.startRun:
          gLogger.error("Your end run '%d' should be more than your start run '%d'!" % (self.endRun, self.startRun))
          raise ValueError("Error setting start or end run")

      if self.startRun:
        bkQuery['StartRun'] = self.startRun
      if self.endRun:
        bkQuery['EndRun'] = self.endRun

      if self.runsList:
        bkQuery['RunNumbers'] = self.runsList.replace(',', ';;;').replace(' ', '')

      if self.visibility:
        bkQuery['Visible'] = self.visibility

    elif mode.lower() == 'frompreviousprod':
      bkQuery = {'FileType': ';;;'.join(fileType).replace(' ', ''),
                 'ProductionID': int(previousProdID)}

      if self.eventType:
        bkQuery['EventType'] = str(self.eventType)

      if self.dqFlag:
        bkQuery['DataQualityFlag'] = self.dqFlag.replace(',', ';;;').replace(' ', '')

    return bkQuery

  ################################################################################
  # properties

  def set_stepsList(self, value):
    listInt = []
    i = 0
    while True:
      try:
        listInt.append(int(value[i]))
        i = i + 1
      except ValueError:
        break
      except IndexError:
        break
    self._stepsList = listInt

  def get_stepsList(self):
    return self._stepsList
  stepsList = property(get_stepsList, set_stepsList)

  def set_startRun(self, value):
    if isinstance(value, six.string_types):
      value = int(value)
    if value < 0:
      raise ValueError("startRun can not be negative")
    self._startRun = value

  def get_startRun(self):
    return self._startRun
  startRun = property(get_startRun, set_startRun)

  def set_endRun(self, value):
    if isinstance(value, six.string_types):
      value = int(value)
    if value < 0:
      raise ValueError("endRun can not be negative")
    self._endRun = value

  def get_endRun(self):
    return self._endRun
  endRun = property(get_endRun, set_endRun)

  def set_requestID(self, value):
    if value == '':
      value = 0
    if isinstance(value, six.string_types):
      value = int(value)
    if value < 0:
      raise ValueError("requestID can not be negative")
    self._requestID = value

  def get_requestID(self):
    return self._requestID
  requestID = property(get_requestID, set_requestID)

  def set_parentRequestID(self, value):
    if value == '':
      value = 0
    if isinstance(value, six.string_types):
      value = int(value)
    if value < 0:
      raise ValueError("parentRequestID can not be negative")
    self._parentRequestID = value

  def get_parentRequestID(self):
    return self._parentRequestID
  parentRequestID = property(get_parentRequestID, set_parentRequestID)

  def set_bkFileType(self, value):
    if isinstance(value, str):
      value = value.replace(' ', '').split(',')
    self._bkFileType = value

  def get_bkFileType(self):
    return self._bkFileType
  bkFileType = property(get_bkFileType, set_bkFileType)

#############################################################################


def _splitIntoProductionSteps(step):
  """Given a list of bookkeeping steps, produce production steps."""
  prodSteps = []

  if len(step['fileTypesIn']) <= 1:
    prodSteps.append(step)
  else:
    if set(step['fileTypesOut']) > set(step['fileTypesIn']):
      raise ValueError("Step outputs %s are not part of the inputs %s...?" % (str(step['fileTypesOut']),
                                                                              str(step['fileTypesIn'])))
    for outputTypes in step['fileTypesOut']:
      prodStep = copy.deepcopy(step)
      prodStep['fileTypesIn'] = [outputTypes]
      prodStep['fileTypesOut'] = [outputTypes]
      prodStep['prodStepID'] = str(prodStep['StepId']) + str(prodStep['fileTypesIn'])
      prodSteps.append(prodStep)

  return prodSteps

#############################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageHandler

.. module: StorageUsageHandler

:synopsis: Implementation of the Storage Usage service in the DISET framework.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# imports
import six

# from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
# from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB

__RCSID__ = "$Id$"

gStorageUsageDB = False


def initializeStorageUsageHandler(_serviceInfo):
  """handlre initialisation."""
  global gStorageUsageDB
  gStorageUsageDB = StorageUsageDB()
  return S_OK()


class StorageUsageHandler(RequestHandler):
  """
  .. class:: StorageUsageHandler
  """

  types_publishDirectories = [dict]

  @staticmethod
  def export_publishDirectories(directoryDict):
    """export of publishDirectories."""
    return gStorageUsageDB.publishDirectories(directoryDict)

  types_removeDirectory = [[six.string_types, list, tuple]]

  @staticmethod
  def export_removeDirectory(dirPaths):
    """export of removeDirectory."""
    if isinstance(dirPaths, six.string_types):
      dirPaths = (dirPaths, )
    for dirPath in dirPaths:
      result = gStorageUsageDB.removeDirectory(dirPath)
      if not result['OK']:
        gLogger.error("Could not delete directory", "%s : %s" % (dirPath, result['Message']))
        return result
    return S_OK()

  types_removeDirFromSe_Usage = []

  @staticmethod
  def export_removeDirFromSe_Usage(dirPaths):
    """Exports the method to remove entries from the se_Usage table."""
    return gStorageUsageDB.removeDirFromSe_Usage(dirPaths)

  types_removeDirFromProblematicDirs = []

  @staticmethod
  def export_removeDirFromProblematicDirs(dirPaths):
    """Exports the method to remove entries from the problematicDirs table."""
    return gStorageUsageDB.removeDirFromProblematicDirs(dirPaths)

  ##################################################################
  #
  # These are the methods for monitoring the usage
  #

  types_getStorageSummary = []

  @staticmethod
  def export_getStorageSummary(directory='', filetype='', production='', sites=None):
    """Retieve a summary for the storage usage."""
    sites = sites if sites else []
    return gStorageUsageDB.getStorageSummary(directory, filetype, production, sites)

  types_getStorageDirectorySummary = []

  @staticmethod
  def export_getStorageDirectorySummary(directory='', filetype='', production='', sites=None):
    """Retieve a directory summary for the storage usage."""
    sites = sites if sites else []
    result = gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, sites)
    if not result['OK']:
      return result
    dl = []
    for dirPath in result['Value']:
      dl.append((dirPath, result['Value'][dirPath]['Size'], result['Value'][dirPath]['Files']))
    return S_OK(dl)

  types_getStorageDirectoryData = []

  @staticmethod
  def export_getStorageDirectoryData(directory='', filetype='', production='', sites=None):
    """Retrieve a directory summary for the storage usage."""
    sites = sites if sites else []
    return gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, sites)

  types_getStorageDirectories = []

  @staticmethod
  def export_getStorageDirectories(directory='', filetype='', production='', sites=None):
    """Retrieve the directories for the supplied selection."""
    sites = sites if sites else []
    return gStorageUsageDB.getStorageDirectories(directory, filetype, production, sites)

  types_getStorageDirectorySummaryWeb = []

  @staticmethod
  def export_getStorageDirectorySummaryWeb(selectDict, sortList, startItem, maxItems):
    """Get the summary of the directory storage summary."""
    resultDict = {}
    # Sorting instructions. Only one for the moment.
    directory = ''
    if "Directory" in selectDict:
      directory = selectDict['Directory']
    filetype = ''
    if "FileType" in selectDict:
      filetype = selectDict['FileType']
    production = ''
    if "Production" in selectDict:
      production = selectDict['Production']
    ses = []
    if "SEs" in selectDict:
      ses = selectDict['SEs']

    res = gStorageUsageDB.getStorageDirectorySummary(directory, filetype, production, ses)
    if not res['OK']:
      gLogger.error("StorageUsageHandler.getStorageDirectorySummaryWeb: Failed to obtain directory summary.",
                    res['Message'])
      return res
    dirList = res['Value']
    dirList = [(path, dirList[path]['Size'], dirList[path]['Files']) for path in dirList]

    nDirs = len(dirList)
    resultDict['TotalRecords'] = nDirs
    if nDirs == 0:
      return S_OK(resultDict)
    iniDir = startItem
    lastDir = iniDir + maxItems
    if iniDir >= nDirs:
      return S_ERROR('Item number out of range')
    if lastDir > nDirs:
      lastDir = nDirs

    # prepare the extras count
    res = gStorageUsageDB.getStorageSummary(directory, filetype, production, ses)
    if not res['OK']:
      gLogger.error("StorageUsageHandler.getStorageDirectorySummaryWeb: Failed to obtain usage summary.",
                    res['Message'])
      return res
    resultDict['Extras'] = res['Value']

    # prepare the standard structure now
    resultDict['ParameterNames'] = ['Directory Path', 'Size', 'Files']
    resultDict['Records'] = dirList[iniDir:lastDir]
    return S_OK(resultDict)

  types_getStorageElementSelection = []

  @staticmethod
  def export_getStorageElementSelection():
    """Retrieve the possible selections."""
    return gStorageUsageDB.getStorageElementSelection()

  types_getUserStorageUsage = []

  @staticmethod
  def export_getUserStorageUsage(userName=False):
    """Retrieve a summary of the user usage."""
    return gStorageUsageDB.getUserStorageUsage(userName)

  types_getUserSummaryPerSE = []

  @staticmethod
  def export_getUserSummaryPerSE(userName=False):
    """Retrieve a summary of the user usage per SE."""
    return gStorageUsageDB.getUserSummaryPerSE(userName)

  types_getDirectorySummaryPerSE = []

  @staticmethod
  def export_getDirectorySummaryPerSE(directory):
    """Retrieve a summary (total files and total size) for a given directory,
    grouped by storage element."""
    return gStorageUsageDB.getDirectorySummaryPerSE(directory)

  types_getRunSummaryPerSE = []

  @staticmethod
  def export_getRunSummaryPerSE(run):
    """Retrieve a summary (total files and total size) for a given run, grouped
    by storage element."""
    return gStorageUsageDB.getRunSummaryPerSE(run)

  types_getIDs = []

  @staticmethod
  def export_getIDs(dirList):
    """Check if the directories exist in the su_Directory table and if yes
    returns the IDs."""
    return gStorageUsageDB.getIDs(dirList)

  types_getAllReplicasInFC = []

  @staticmethod
  def export_getAllReplicasInFC(path):
    """Export the DB method to query the su_seUsage table to get all the
    entries relative to a given path registered in the FC.

    Returns for every replica the SE, the update, the files and the size
    """
    return gStorageUsageDB.getAllReplicasInFC(path)

  ####
  # Catalog
  ####

  types_getSummary = [six.string_types]

  @staticmethod
  def export_getSummary(path, fileType=False, production=False):
    """export of getSummary."""
    return gStorageUsageDB.getSummary(path, fileType, production)

  types_getUserSummary = []

  @staticmethod
  def export_getUserSummary(userName=False):
    """export of getUserSummary."""
    return gStorageUsageDB.getUserSummary(userName)

  ####
  # Purge
  ####

  types_purgeOutdatedEntries = [six.string_types, int]

  @staticmethod
  def export_purgeOutdatedEntries(rootDir, outdatedSeconds, preserveDirsList=None):
    """Purge entries that haven't been updated in the last outdated seconds."""
    preserveDirsList = preserveDirsList if preserveDirsList else []
    return gStorageUsageDB.purgeOutdatedEntries(rootDir, outdatedSeconds, preserveDirsList)

  ###
  # methods to deal with problematicDirs directory: problematicDirs
  ###
  types_publishToProblematicDirs = []

  @staticmethod
  def export_publishToProblematicDirs(directoryDict):
    """Export the publishToProblematicDirs DB method, which inserts/updates row
    into the  problematicDirs."""
    return gStorageUsageDB.publishToProblematicDirs(directoryDict)

  types_getProblematicDirsSummary = []

  @staticmethod
  def export_getProblematicDirsSummary(site, problem=False):
    """Exports the getProblematicDirsSummary method: returns a list of
    directories from the problematicDirs table, that have some inconsistency
    between the SE dumps and the LFC."""
    return gStorageUsageDB.getProblematicDirsSummary(site, problem)

  types_removeAllFromProblematicDirs = []

  @staticmethod
  def export_removeAllFromProblematicDirs(site=False):
    """Exports the removeAllFromProblematicDirs method: delete all entries from
    problematicDirs table for a give site (optional argument)"""
    return gStorageUsageDB.removeAllFromProblematicDirs(site)

  ###
  # methods to deal with se_Usage table
  ###
  types_publishToSEReplicas = []

  @staticmethod
  def export_publishToSEReplicas(directoryDict):
    """Export the publishToSEReplicas DB method, which inserts/updates replicas
    on the SE to the se_Usage table."""
    return gStorageUsageDB.publishToSEReplicas(directoryDict)

  ###
  # methods to deal with se_STSummary table
  ###

  types_publishTose_STSummary = []

  @staticmethod
  def export_publishTose_STSummary(site, spaceToken, totalSize, totalFiles, StorageDumpUpdate):
    """Export the publishTose_STSummary DB method, which inserts/updates the
    reports of total files and total size from the storage dumps to the
    se_STSummary table."""
    return gStorageUsageDB.publishTose_STSummary(site, spaceToken, totalSize, totalFiles, StorageDumpUpdate)

  types_getSTSummary = []

  @staticmethod
  def export_getSTSummary(site, spaceToken=False):
    """Exports getSTSummary method: returns a summary of the used space for the
    given site, based on the storage dumps provided by sites."""
    return gStorageUsageDB.getSTSummary(site, spaceToken)

  types_removeSTSummary = []

  @staticmethod
  def export_removeSTSummary(site, spaceToken=False):
    """Exports removeSTSummary method: removes from the se_STSummary table all
    entries relative to the given site and (optionally ) space token."""
    return gStorageUsageDB.removeSTSummary(site, spaceToken)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""A module defining the state machine for the Productions."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import State, StateMachine

__RCSID__ = "$Id$"


class LHCbStateMachine(StateMachine):
  """Just redefining one method."""

  def setState(self, candidateState):
    """Makes sure the state is either None or known to the machine, and that it
    is a valid state to move into.

    Final states are also checked. This is a re-definition of original
    one that wasn't making these checks
    """

    if candidateState == self.state:
      return S_OK(candidateState)

    if candidateState is None:
      self.state = candidateState
    elif candidateState in self.states:
      if not self.states[self.state].stateMap:
        gLogger.warn("Final state, won't move")
        return S_OK(self.state)
      if candidateState not in self.states[self.state].stateMap:
        gLogger.warn("Can't move from %s to %s, choosing a good one" % (self.state, candidateState))
      nextState = self.getNextState(candidateState)
      if not nextState['OK']:
        return nextState
      nextState = nextState['Value']
      # If the StateMachine does not accept the candidate, return error message
      self.state = nextState
    else:
      return S_ERROR("%s is not a valid state" % candidateState)

    return S_OK(nextState)


class ProductionsStateMachine(LHCbStateMachine):
  """PMS (Production Management System, not what google thinks!) implementation
  of the state machine."""

  def __init__(self, state):
    """c'tor Defines the state machine transactions."""

    super(ProductionsStateMachine, self).__init__(state)

    # Current states
    self.states = {'Cleaned': State(15),  # final state
                   'Cleaning': State(14, ['Cleaned', 'TransformationCleaned'],
                                     defState='Cleaning'),
                   'Completing': State(13, ['Validating'],
                                       defState='Completing'),
                   'Stopped': State(12, ['Active'],
                                    defState='Stopped'),
                   'TransformationCleaned': State(16, ['Archived', 'Cleaned'],
                                                  defState='Archived'),
                   'Archived': State(11),  # final state
                   'Completed': State(10, ['TransformationCleaned', 'Archived'],
                                      defState='Completed'),
                   'WaitingIntegrity': State(9, ['ValidatedOutput'],
                                             defState='WaitingIntegrity'),
                   'ValidatedOutput': State(8, ['Active', 'Completed'],
                                            defState='ValidatedOutput'),
                   'ValidatingOutput': State(7, ['Active', 'ValidatedOutput', 'WaitingIntegrity'],
                                             defState='ValidatedOutput'),
                   'RemovedFiles': State(6, ['Completed'],
                                         defState='RemovedFiles'),
                   'RemovingFiles': State(5, ['RemovedFiles'],
                                          defState='RemovingFiles'),
                   'ValidatingInput': State(4, ['Active', 'RemovingFiles'],
                                            defState='ValidatingInput'),
                   'Flush': State(3, ['Active'],
                                  defState='Active'),
                   'Idle': State(2, ['Active', 'Stopped', 'ValidatingInput',
                                     'ValidatingOutput', 'Completed', 'Cleaning', 'Testing'],
                                 defState='Idle'),
                   'Active': State(1, ['Flush', 'Idle', 'Stopped', 'Completing', 'ValidatingInput',
                                       'ValidatingOutput', 'TransformationCleaned'],
                                   defState='Flush'),
                   'Testing': State(17, ['Idle'], defState='Testing'),
                   'New': State(0, ['Active', 'Testing', 'Cleaning'],
                                defState='New')  # initial state
                   }

    # NEW states proposal
    # self.states = {
    #     'Cleaned': State(10), # final state
    #     'Cleaning': State(9, ['Cleaned']),
    #     'Completing': State(8, ['Validating', 'Cleaning'], defState = 'Validating'),
    #     'TransformationCleaned': State(16, ['Archived', 'Cleaned'], defState = 'Archived'),
    #     'Stopped': State(7, ['Active', 'Flush', 'Cleaning'], defState = 'Active'),
    #     'Archived': State(6),  # final state
    #     'Completed': State(5, ['Archived'], defState = 'Archived'),
    #     'Validating': State(4, ['Active', 'Completed', 'Cleaning'], defState = 'Completed'),
    #     # Idle to Completed? Should it be that when setting a prod to completed it actually goes to validating?
    #     'Idle': State(3, ['Active', 'Validating', 'Cleaning'], defState = 'Active'),
    #     'Flush': State(2, ['Active', 'Cleaning'], defState = 'Active'),
    #     'Active': State(1, ['Flush', 'Idle', 'Stopped', 'Completing', 'Validating', 'Cleaning'], defState = 'Flush'),
    #     'New': State(0, ['Active', 'Cleaning'], defState = 'Active')  # initial state
    # }



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get the list of all the user files."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript, Script

  days = 0
  months = 0
  years = 0
  depth = 1
  wildcard = '*'
  dmScript = DMScript()
  dmScript.registerNamespaceSwitches()
  Script.registerSwitch("", "Days=", "Match files older than number of days [%s]" % days)
  Script.registerSwitch("", "Months=", "Match files older than number of months [%s]" % months)
  Script.registerSwitch("", "Years=", "Match files older than number of years [%s]" % years)
  Script.registerSwitch("", "Wildcard=", "Wildcard for matching filenames [%s]" % wildcard)
  Script.registerSwitch('', 'Output', 'Write list to an output file')
  Script.registerSwitch("", "EmptyDirs", "Create a list of empty directories")
  Script.registerSwitch("", "Depth=", "Depth to which recursively browse (default = %d)" % depth)
  Script.registerSwitch("r", "Recursive", "Set depth to infinite")
  Script.registerSwitch('', 'NoDirectories', 'Only print out only files, not subdirectories')
  dmScript.registerBKSwitches()

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName, ]))

  Script.parseCommandLine(ignoreErrors=False)
  from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeListDirectory
  from DIRAC import exit

  exit(executeListDirectory(dmScript, days, months, years, wildcard, depth))


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for:
LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter.

DataStoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

import unittest
import math
import operator

from decimal import Decimal

import mock

from PIL import Image
from functools import reduce


def compare(file1Path, file2Path):
  """Function used to compare two plots.

  returns 0.0 if both are identical
  """

  # Crops image to remove the "Generated on xxxx UTC" string
  image1 = Image.open(file1Path).crop((0, 0, 800, 570))
  image2 = Image.open(file2Path).crop((0, 0, 800, 570))
  h1 = image1.histogram()
  h2 = image2.histogram()
  rms = math.sqrt(reduce(operator.add,
                         map(lambda a, b: (a - b) ** 2, h1, h2)) / len(h1))
  return rms

# ...............................................................................


class DataStoragePlotterTestCase(unittest.TestCase):
  """DataStoragePlotterTestCase."""

  moduleTested = None
  classsTested = None

  def mockModuleTested(self, moduleTested):
    """Used to not redo the mocking done on the parent class ( if any )"""

    # Tries to get the mocks of the parent TestCases ( if any )
    for baseClass in self.__class__.__bases__:
      try:
        # pylint: disable=no-member
        moduleTested = baseClass.mockModuleTested(moduleTested)
      except TypeError:
        continue

    # And then makes its own mock
    class MockDataStorage(object):
      # pylint: disable=missing-docstring,no-init
      definitionKeyFields = ('DataType', 'Activity', 'FileType', 'Production',
                             'ProcessingPass', 'Conditions', 'EventType',
                             'StorageElement')

    moduleTested.DataStorage = mock.Mock(return_value=MockDataStorage())

    return moduleTested

  def setUp(self):
    """Setup the test case."""

    import LHCbDIRAC.AccountingSystem.private.Plotters.DataStoragePlotter as moduleTested

    self.moduleTested = self.mockModuleTested(moduleTested)
    self.classsTested = self.moduleTested.DataStoragePlotter

  def tearDown(self):
    """Tear down the test case."""

    del self.moduleTested
    del self.classsTested

# ...............................................................................


class DataStoragePlotterUnitTest(DataStoragePlotterTestCase):
  """DataStoragePlotterUnitTest.

  <constructor>
   - test_instantiate
  <class variables>
   - test_typeName
   - test_typeKeyFields
   - test_noSEtypeKeyFields
   - test_noSEGrouping
   - test_reportCatalogSpaceName
   - test_reportCatalogFilesName
   - test_reportPhysicalSpaceName
   - test_reportPhysicalFilesName
  <methods>
   - test_reportCatalogSpace
   - test_reportCatalogFiles
   - test_reportPhysicalSpace
   - test_reportPhysicalFiles
   - test_plotCatalogSpace
   - test_plotCatalogFiles
   - test_plotPhysicalSpace
   - test_plotPhysicalFiles
  """

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""
    obj = self.classsTested(None, None)
    self.assertEqual('DataStoragePlotter', obj.__class__.__name__)

  def test_typeName(self):
    """test the class variable "_typeName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._typeName, "DataStorage")

#   def test_typeKeyFields( self ):
#     ''' test the class variable "_typeKeyFields"
#     '''
#     obj = self.classsTested( None, None )
#     self.assertEqual( obj._typeKeyFields, [ 'DataType', 'Activity', 'FileType',
#                                             'Production', 'ProcessingPass',
#                                             'Conditions', 'EventType', 'StorageElement' ] )

  def test_reportCatalogSpaceName(self):
    """test the class variable "_reportCatalogSpaceName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportCatalogSpaceName, "LFN size")

  def test_reportCatalogFilesName(self):
    """test the class variable "_reportCatalogFilesName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportCatalogFilesName, "LFN files")

  def test_reportPhysicalSpaceName(self):
    """test the class variable "_reportPhysicalSpaceName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportPhysicalSpaceName, "PFN size")

  def test_reportPhysicalFilesName(self):
    """test the class variable "_reportPhysicalFilesName"."""
    obj = self.classsTested(None, None)
    self.assertEqual(obj._reportPhysicalFilesName, "PFN files")

  def test_reportCatalogSpace(self):
    """test the method "_reportCatalogSpace"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    res = obj._reportCatalogSpace({'grouping': 'StorageElement'})
    self.assertEqual(res['OK'], False, msg='Rejected StorageElement grouping')
    self.assertEqual(res['Message'], 'Grouping by storage element when requesting lfn info makes no sense')

    res = obj._reportCatalogSpace({'grouping': 'NextToABeer',
                                   'groupingFields': [0, ['mehh'], 'blah'],
                                   'startTime': 'startTime',
                                   'endTime': 'endTime',
                                   'condDict': {}
                                   })
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportCatalogSpace({'grouping': 'NextToABeer',
                                   'groupingFields': [0, ['mehh'], 'blah'],
                                   'startTime': 'startTime',
                                   'endTime': 'endTime',
                                   'condDict': {}
                                   })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'MB',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('Full stream', 1355616000, 86400, Decimal('935388524246.91630989384787')),
                  ('Full stream', 1355702400, 86400, Decimal('843844487074.82197482051816')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportCatalogSpace({'grouping': 'EventType',
                                   'groupingFields': ('%s', ['EventType']),
                                   'startTime': 1355663249.0,
                                   'endTime': 1355749690.0,
                                   'condDict': {'EventType': 'Full stream'}
                                   })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'Full stream': {1355616000: 935.38852424691629,
                                                                      1355702400: 843.84448707482204
                                                                      }
                                                      },
                                    'data': {'Full stream': {1355616000: 935388.52424691629,
                                                             1355702400: 843844.48707482207
                                                             }
                                             },
                                    'unit': 'GB',
                                    'granularity': 86400
                                    })

  def test_reportCatalogFiles(self):
    """test the method "_reportCatalogFiles"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    res = obj._reportCatalogFiles({'grouping': 'StorageElement'})
    self.assertEqual(res['OK'], False)
    self.assertEqual(res['Message'], 'Grouping by storage element when requesting lfn info makes no sense')

    res = obj._reportCatalogFiles({'grouping': 'NextToABeer',
                                   'groupingFields': [0, ['mehh'], 'blah'],
                                   'startTime': 'startTime',
                                   'endTime': 'endTime',
                                   'condDict': {}
                                   })
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportCatalogFiles({'grouping': 'NextToABeer',
                                   'groupingFields': [0, ['mehh'], 'blah'],
                                   'startTime': 'startTime',
                                   'endTime': 'endTime',
                                   'condDict': {}
                                   })
    self.assertEqual(res['OK'], True, msg='Expected S_OK')
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'files',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('Full stream', 1355616000, 86400, Decimal('420.47885754501202')),
                  ('Full stream', 1355702400, 86400, Decimal('380.35170637810842')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportCatalogFiles({'grouping': 'EventType',
                                   'groupingFields': ('%s', ['EventType']),
                                   'startTime': 1355663249.0,
                                   'endTime': 1355749690.0,
                                   'condDict': {'EventType': 'Full stream'}
                                   })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'Full stream': {1355616000: 420.47885754501203,
                                                                      1355702400: 380.35170637810842}
                                                      },
                                    'data': {'Full stream': {1355616000: 420.47885754501203,
                                                             1355702400: 380.35170637810842}
                                             },
                                    'unit': 'files',
                                    'granularity': 86400
                                    })

  def test_reportPhysicalSpace(self):
    """test the method "_reportPhysicalSpace"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    res = obj._reportPhysicalSpace({'groupingFields': [0, ['mehh'], 'blah'],
                                    'startTime': 'startTime',
                                    'endTime': 'endTime',
                                    'condDict': {}
                                    })
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportPhysicalSpace({'grouping': 'NextToABeer',
                                    'groupingFields': [0, ['mehh'], 'blah'],
                                    'startTime': 'startTime',
                                    'endTime': 'endTime',
                                    'condDict': {}
                                    })
    self.assertEqual(res['OK'], True, msg='Expected S_OK')
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'MB',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('Full stream', 1355616000, 86400, Decimal('14754501.202')),
                  ('Full stream', 1355702400, 86400, Decimal('15237810.842')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportPhysicalSpace({'grouping': 'EventType',
                                    'groupingFields': ('%s', ['EventType']),
                                    'startTime': 1355663249.0,
                                    'endTime': 1355749690.0,
                                    'condDict': {'EventType': 'Full stream'}
                                    })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'Full stream': {1355616000: 14.754501202,
                                                                      1355702400: 15.237810842}
                                                      },
                                    'data': {'Full stream': {1355616000: 14.754501202,
                                                             1355702400: 15.237810842}
                                             },
                                    'unit': 'MB',
                                    'granularity': 86400
                                    })

  def test_reportPhysicalFiles(self):
    """test the method "_reportPhysicalFiles"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    res = obj._reportPhysicalFiles({'groupingFields': [0, ['mehh'], 'blah'],
                                    'startTime': 'startTime',
                                    'endTime': 'endTime',
                                    'condDict': {}
                                    })
#     self.assertEqual( res[ 'OK' ], False )
#     self.assertEqual( res[ 'Message' ], 'No connection' )

    # Changed mocked to run over different lines of code
    mockAccountingDB._getConnection.return_value = {'OK': True, 'Value': []}
    res = obj._reportPhysicalFiles({'grouping': 'NextToABeer',
                                    'groupingFields': [0, ['mehh'], 'blah'],
                                    'startTime': 'startTime',
                                    'endTime': 'endTime',
                                    'condDict': {}
                                    })
    self.assertEqual(res['OK'], True, msg='Expected S_OK')
    self.assertEqual(res['Value'], {'graphDataDict': {},
                                    'data': {},
                                    'unit': 'files',
                                    'granularity': 'BucketLength'
                                    })

    mockedData = (('Full stream', 1355616000, 86400, Decimal('42.47885754501202')),
                  ('Full stream', 1355702400, 86400, Decimal('38.35170637810842')))
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': mockedData}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

    res = obj._reportPhysicalFiles({'grouping': 'EventType',
                                    'groupingFields': ('%s', ['EventType']),
                                    'startTime': 1355663249.0,
                                    'endTime': 1355749690.0,
                                    'condDict': {'EventType': 'Full stream'}
                                    })
    self.assertEqual(res['OK'], True)
    self.assertEqual(res['Value'], {'graphDataDict': {'Full stream': {1355616000: 42.47885754501202,
                                                                      1355702400: 38.35170637810842}
                                                      },
                                    'data': {'Full stream': {1355616000: 42.47885754501202,
                                                             1355702400: 38.35170637810842}
                                             },
                                    'unit': 'files',
                                    'granularity': 86400
                                    })

# ...............................................................................


class DataStoragePlotterUnitTestCrashes(DataStoragePlotterTestCase):
  """DataStoragePlotterUnitTestCrashes.

  <constructor>
   - test_instantiate
  <class variables>
  <methods>
   - test_reportCatalogSpace
   - test_reportCatalogFiles
   - test_reportPhysicalSpace
   - test_reportPhysicalFiles
   - test_plotCatalogSpace
   - test_plotCatalogFiles
   - test_plotPhysicalSpace
   - test_plotPhysicalFiles
  """

  def test_instantiate(self):
    """test the constructor."""

    self.assertRaises(TypeError, self.classsTested)
    self.assertRaises(TypeError, self.classsTested, None)
    self.assertRaises(TypeError, self.classsTested, None, None, None, None)

    self.assertRaises(TypeError, self.classsTested, extraArgs=None)
    self.assertRaises(TypeError, self.classsTested, None, extraArgs=None)
    self.assertRaises(TypeError, self.classsTested, None, None, None, extraArgs=None)

  def test_reportCatalogSpace(self):
    """test the method "_reportCatalogSpace"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportCatalogSpace, {})
    self.assertRaises(KeyError, obj._reportCatalogSpace, {'grouping': 1})
    self.assertRaises(IndexError, obj._reportCatalogSpace, {'grouping': 1,
                                                            'groupingFields': []})
    self.assertRaises(TypeError, obj._reportCatalogSpace, {'grouping': 1,
                                                           'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportCatalogSpace, {'grouping': 1,
                                                           'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportCatalogSpace, {'grouping': 1,
                                                           'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportCatalogSpace, {'grouping': 1,
                                                          'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportCatalogSpace, {'grouping': 1,
                                                          'groupingFields': ['1', [2, 2]],
                                                          'startTime': None})
    self.assertRaises(KeyError, obj._reportCatalogSpace, {'grouping': 1,
                                                          'groupingFields': ['1', [2, 2]],
                                                          'startTime': None,
                                                          'endTime': None})
    self.assertRaises(TypeError, obj._reportCatalogSpace, {'grouping': 1,
                                                           'groupingFields': ['1', [2, 2]],
                                                           'startTime': None,
                                                           'endTime': None,
                                                           'condDict': None})

  def test_reportCatalogFiles(self):
    """test the method "_reportCatalogFiles"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportCatalogFiles, {})
    self.assertRaises(KeyError, obj._reportCatalogFiles, {'grouping': 1})
    self.assertRaises(IndexError, obj._reportCatalogFiles, {'grouping': 1,
                                                            'groupingFields': []})
    self.assertRaises(TypeError, obj._reportCatalogFiles, {'grouping': 1,
                                                           'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportCatalogFiles, {'grouping': 1,
                                                           'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportCatalogFiles, {'grouping': 1,
                                                           'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportCatalogFiles, {'grouping': 1,
                                                          'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportCatalogFiles, {'grouping': 1,
                                                          'groupingFields': ['1', [2, 2]],
                                                          'startTime': None})
    self.assertRaises(KeyError, obj._reportCatalogFiles, {'grouping': 1,
                                                          'groupingFields': ['1', [2, 2]],
                                                          'startTime': None,
                                                          'endTime': None})
    self.assertRaises(TypeError, obj._reportCatalogFiles, {'grouping': 1,
                                                           'groupingFields': ['1', [2, 2]],
                                                           'startTime': None,
                                                           'endTime': None,
                                                           'condDict': None})

  def test_reportPhysicalSpace(self):
    """test the method "_reportPhysicalSpace"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportPhysicalSpace, {})
    self.assertRaises(IndexError, obj._reportPhysicalSpace, {'groupingFields': []})
    self.assertRaises(TypeError, obj._reportPhysicalSpace, {'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportPhysicalSpace, {'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportPhysicalSpace, {'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportPhysicalSpace, {'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportPhysicalSpace, {'groupingFields': ['1', [2, 2]],
                                                           'startTime': None})
    self.assertRaises(KeyError, obj._reportPhysicalSpace, {'groupingFields': ['1', [2, 2]],
                                                           'startTime': None,
                                                           'endTime': None})
    self.assertRaises(TypeError, obj._reportPhysicalSpace, {'groupingFields': ['1', [2, 2]],
                                                            'startTime': None,
                                                            'endTime': None,
                                                            'condDict': None})

  def test_reportPhysicalFiles(self):
    """test the method "_reportPhysicalFiles"."""

    mockAccountingDB = mock.Mock()
    mockAccountingDB._getConnection.return_value = {'OK': False, 'Message': 'No connection'}
    mockAccountingDB.retrieveBucketedData.return_value = {'OK': True, 'Value': []}
    mockAccountingDB.calculateBucketLengthForTime.return_value = 'BucketLength'
    obj = self.classsTested(mockAccountingDB, None)

    self.assertRaises(KeyError, obj._reportPhysicalFiles, {})
    self.assertRaises(IndexError, obj._reportPhysicalFiles, {'groupingFields': []})
    self.assertRaises(TypeError, obj._reportPhysicalFiles, {'groupingFields': [1, 2]})
    self.assertRaises(TypeError, obj._reportPhysicalFiles, {'groupingFields': [1, [2]]})
    self.assertRaises(TypeError, obj._reportPhysicalFiles, {'groupingFields': ['1', '2']})
    self.assertRaises(KeyError, obj._reportPhysicalFiles, {'groupingFields': ['1', [2]]})
    self.assertRaises(KeyError, obj._reportPhysicalFiles, {'groupingFields': ['1', [2, 2]],
                                                           'startTime': None})
    self.assertRaises(KeyError, obj._reportPhysicalFiles, {'groupingFields': ['1', [2, 2]],
                                                           'startTime': None,
                                                           'endTime': None})
    self.assertRaises(TypeError, obj._reportPhysicalFiles, {'groupingFields': ['1', [2, 2]],
                                                            'startTime': None,
                                                            'endTime': None,
                                                            'condDict': None})

  def test_plotCatalogSpace(self):
    """test the method "_plotCatalogSpace"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotCatalogSpace, None, None, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {}, None, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotCatalogSpace, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {'startTime': 'startTime',
                                                        'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {'startTime': 'startTime',
                                                        'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {'startTime': 'startTime',
                                                        'endTime': 'endTime',
                                                        'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotCatalogSpace, {'startTime': 'startTime',
                                                        'endTime': 'endTime',
                                                        'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

  def test_plotCatalogFiles(self):
    """test the method "_plotCatalogFiles"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotCatalogFiles, None, None, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {}, None, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotCatalogFiles, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {'startTime': 'startTime',
                                                        'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {'startTime': 'startTime',
                                                        'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {'startTime': 'startTime',
                                                        'endTime': 'endTime',
                                                        'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotCatalogFiles, {'startTime': 'startTime',
                                                        'endTime': 'endTime',
                                                        'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

  def test_plotPhysicalSpace(self):
    """test the method "_plotPhysicalSpace"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotPhysicalSpace, None, None, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {}, None, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotPhysicalSpace, {'startTime': 'startTime',
                                                          'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {'startTime': 'startTime',
                                                         'endTime': 'endTime',
                                                         'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotPhysicalSpace, {'startTime': 'startTime',
                                                         'endTime': 'endTime',
                                                         'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

  def test_plotPhysicalFiles(self):
    """test the method "_plotPhysicalFiles"."""

    obj = self.classsTested(None, None)
    self.assertRaises(TypeError, obj._plotPhysicalFiles, None, None, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {}, None, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {'startTime': 'startTime'},
                      None, None)
    self.assertRaises(TypeError, obj._plotPhysicalFiles, {'startTime': 'startTime',
                                                          'endTime': 'endTime'},
                      None, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      {}, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {'startTime': 'startTime',
                                                         'endTime': 'endTime'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {'startTime': 'startTime',
                                                         'endTime': 'endTime',
                                                         'grouping': 'grouping'},
                      {'granularity': 'granularity'}, None)
    self.assertRaises(KeyError, obj._plotPhysicalFiles, {'startTime': 'startTime',
                                                         'endTime': 'endTime',
                                                         'grouping': 'grouping'},
                      {'granularity': 'granularity',
                       'graphDataDict': 'graphDataDict'}, None)

################################################################################
#############################################################################


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterTestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterTestCase))
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DataStoragePlotterUnitTestCrashes))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""List simulation conditions from the Bookkeeping."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ...' % Script.scriptName]))
  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  exitCode = 0

  res = BookkeepingClient().getSimConditions()
  if res['OK']:
    dbresult = res['Value']
    for record in dbresult:
      print('SimId: ' + str(record[0]).ljust(10))
      print('  SimDescription: ' + str(record[1]).ljust(10))
      print('  BeamCond: ' + str(record[2]).ljust(10))
      print('  BeamEnergy: ' + str(record[3]).ljust(10))
      print('  Generator: ' + str(record[4]).ljust(10))
      print('  MagneticField: ' + str(record[5]).ljust(10))
      print('  DetectorCond: ' + str(record[6]).ljust(10))
      print('  Luminosity: ' + str(record[7]).ljust(10))
      print('  G4settings: ' + str(record[8]).ljust(10))

  else:
    print('ERROR:', res['Message'])
    exitCode = 1

  DIRAC.exit(exitCode)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" JSON encoder - slighltly modified version for LHCbDIRAC purposes
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import datetime


__RCSID__ = "$Id$"


class JSONDateTimeEncoder(json.JSONEncoder):
  """encoder of datetime objects."""

  def default(self, obj):  # pylint: disable=method-hidden
    if isinstance(obj, datetime.datetime):
      return {'dt': [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]}
    elif isinstance(obj, datetime.date):
      return {'d': [obj.year, obj.month, obj.day]}
    elif isinstance(obj, datetime.time):
      return {'t': [obj.hour, obj.minute, obj.second]}

    return super(JSONDateTimeEncoder, self).default(obj)


class JSONDateTimeDecoder(json.JSONDecoder):
  """decoder of datetime objects."""

  def __init__(self, *args, **kwargs):
    json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

  def object_hook(self, obj):  # pylint: disable=method-hidden
    if 'dt' in obj:
      obj = datetime.datetime(obj['dt'][0], obj['dt'][1], obj['dt'][2], obj['dt'][3], obj['dt'][4], obj['dt'][5])
    elif 'd' in obj:
      obj = datetime.date(obj['d'][0], obj['d'][1], obj['d'][2])
    elif 't' in obj:
      obj = datetime.time(obj['t'][0], obj['t'][1], obj['t'][2])

    return obj


def dumps(obj):
  return json.dumps(obj, cls=JSONDateTimeEncoder)


def loads(obj):
  return json.loads(obj, cls=JSONDateTimeDecoder)


def load(fd):
  return json.load(fd, cls=JSONDateTimeDecoder)
